import base64
import cmath
import codecs
import csv
import hashlib
import itertools
import json
import logging
import math
import os
import pytest
import random
import secrets
import multiprocessing

from collections	import deque

import shamir_mnemonic

from .api		import create, account, path_hardened
from .recovery		import recover, recover_bip39, shannon_entropy, signal_entropy, analyze_entropy
from .recovery.entropy	import fft, ifft, pfft, dft, dft_on_real, dft_to_rms_mags, entropy_bin_dfts, denoise_mags, signal_draw, signal_recover_real, scan_entropy
from .dependency_test	import substitute, nonrandom_bytes, SEED_XMAS, SEED_ONES, SEED_ZERO
from .util		import avg, rms, ordinal, commas, round_onto

log				= logging.getLogger( __package__ )

groups_example			= dict( one = (1,1), two = (1,1), fam = (2,4), fren = (3,5) )

# Disable printing of details unless something goes wrong...
print_NOOP			= lambda *args, **kwds: None
print				= print_NOOP

def noise( mag ):
    return mag * ( random.random() * 2 - 1 )


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_recover():
    details			= create(
        "recovery test", 2, groups_example, SEED_XMAS
    )
    #import json
    #print( json.dumps( details.groups, indent=4 ))
    assert details.groups == {
        "one": (
            1,
            [
                "academic acid acrobat romp chubby client grief judicial pulse domain flip elevator become spirit heat patent hawk remove pickup boring"
            ]
        ),
        "two": (
            1,
            [
                "academic acid beard romp away ancient domain jacket early admit true disaster manual sniff seafood guest stick grumpy blessing unknown"
            ]
        ),
        "fam": (
            2,
            [
                "academic acid ceramic roster density snapshot crush modify born plastic greatest victim merit weapon general cover wits cradle quick emphasis",
                "academic acid ceramic scared brother carve scout stay repeat that fumes tendency junior clay freshman rhyme infant enlarge puny decent",
                "academic acid ceramic shadow class findings zero blessing sidewalk drink jump hormone advocate flip install alpha ugly speak prospect solution",
                "academic acid ceramic sister aluminum obesity blue furl grownup island educate junk traveler listen evidence merit grant python purchase piece"
            ]
        ),
        "fren": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic ranked flame amount",
                "academic acid decision scatter change pleasure dive cricket class impulse lungs hour invasion strike mustang friendly divorce corner penalty fawn",
                "academic acid decision shaft disaster python expand math typical screw rumor research unusual segment install curly debut shadow orange museum",
                "academic acid decision skin browser breathe intimate picture smirk railroad equip spirit nervous capital teaspoon hybrid angel findings hunting similar",
                "academic acid decision snake angel phrase gums response tracks carve secret bucket liquid dictate enemy decrease dance early weapon season"
            ]
        )
    }
    assert recover( details.groups['one'][1] + details.groups['fren'][1][:3] ) == SEED_XMAS

    # Enough correct number of mnemonics must be provided (extras ignored)
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] )
    assert "Wrong number of mnemonics" in str(excinfo.value)

    assert recover( details.groups['one'][1] + details.groups['fren'][1][:4] ) == SEED_XMAS

    # Invalid mnemonic phrases are rejected (one word changed)
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + [
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry laundry"
        ])
    assert "Invalid mnemonic checksum" in str(excinfo.value)

    # Duplicate mnemonics rejected/ignored
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + details.groups['fren'][1][:1] )
    assert "Wrong number of mnemonics" in str(excinfo.value)

    # Mnemonics from another SLIP-39 rejected
    with pytest.raises(shamir_mnemonic.MnemonicError) as excinfo:
        recover( details.groups['one'][1] + details.groups['fren'][1][:2] + [
            "academic acid academic axle crush swing purple violence teacher curly total equation clock mailman display husband tendency smug laundry disaster"
        ])
    assert "Invalid set of mnemonics" in str(excinfo.value)


@substitute( shamir_mnemonic.shamir, 'RANDOM_BYTES', nonrandom_bytes )
def test_recover_bip39():
    """Go through the 3 methods for producing accounts from the same 0xffff...ffff Seed Entropy."""

    # Get BIP-39 Seed generated from Mnemonic Entropy + passphrase
    bip39seed			= recover_bip39( 'zoo ' * 11 + 'wrong' )
    assert codecs.encode( bip39seed, 'hex_codec' ).decode( 'ascii' ) \
        == 'b6a6d8921942dd9806607ebc2750416b289adea669198769f2e15ed926c3aa92bf88ece232317b4ea463e84b0fcd3b53577812ee449ccc448eb45e6f544e25b6'
    details_bip39		= create(
        "bip39 recovery test", 2, groups_example, master_secret=bip39seed,
    )
    #import json
    #print( json.dumps( details_bip39.groups, indent=4 ))
    assert details_bip39.groups == {
        "one": (
            1,
            [
                "academic acid acrobat romp academic angel email prospect endorse strategy debris award strike frost actress facility legend safari pistol"
                " mouse hospital identify unwrap talent entrance trust cause ranked should impulse avoid fangs various radar dilemma indicate says rich work"
                " presence jerky glance hesitate huge depend tension loan tolerate news agree geology phrase random simple finger alarm depart inherit grin"
            ]
        ),
        "two": (
            1,
            [
                "academic acid beard romp acne floral cricket answer debris making decorate square withdraw empty decorate object artwork tracks rocky tolerate"
                " syndrome decorate predator sweater ordinary pecan plastic spew facility predator miracle change solution item lizard testify coal excuse lecture"
                " exercise hamster hand crystal rainbow indicate phantom require satisfy flame acrobat detect closet patent therapy overall muscle spill adjust unhappy"
            ]
        ),
        "fam": (
            2,
            [
                "academic acid ceramic roster acquire again tension ugly edge profile custody geology listen hazard smug branch adequate fishing simple adapt fancy"
                " hour method emperor tactics float quiet location satoshi guilt fantasy royal machine dictate squeeze devote oven eclipse writing level sheriff"
                " teacher purchase building veteran spirit woman realize width vanish scholar jewelry desktop stilt random rhyme debut premium theater",
                "academic acid ceramic scared acid space fantasy breathe true recover privacy tactics boring harvest punish swimming leader talent exchange diet"
                " enforce vanish volume organize coastal emperor change intend club scene intimate upgrade dragon burning lily huge market calcium forecast holiday"
                " merit method type ruler equip retailer pancake paces thorn worthy always story promise clock staff floral smart iris repair",
                "academic acid ceramic shadow acne rumor decent elder aspect lizard obesity friendly regular aircraft beyond military campus employer seafood cover"
                " ivory dough galaxy victim diminish average music cause behavior declare brave toxic visual academic include lilac repair morning rapids building"
                " kernel herald careful helpful move hawk flash glimpse seafood listen writing rocky browser change hybrid diet organize system wrote",
                "academic acid ceramic sister academic both legend raspy pecan mixed broken tenant critical again imply finance pacific single echo capital hesitate"
                " piece disease crush slush belong airline smug voice organize dryer standard emission curious charity swing pitch senior behavior vintage chemical"
                " cage editor rebuild costume adult ancestor erode steady makeup depart carpet level sympathy being soldier glimpse airport picture"
            ]
        ),
        "fren": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic academic"
                " academic academic academic academic academic academic academic aviation endless plastic",
                "academic acid decision scatter acid ugly raspy famous swimming else length gray raspy brother fake aunt auction premium military emphasis perfect"
                " surprise class suitable crunch famous burden military laundry inmate regret elder mixture tenant taught smirk voter process steady artist equip"
                " jury carve acrobat western cylinder gasoline artwork snapshot ancestor object cinema market species platform iris dragon dive medal",
                "academic acid decision shaft acid carbon credit cards rich living humidity peasant source triumph magazine ladle ruin ocean aspect curious round"
                " main evoke deny stadium zero discuss union strike pencil golden silent geology display wrap peanut listen aide learn juice decision plot bike example"
                " obesity ancient square pistol twice sister hour amuse human hobo hospital escape expect wildlife luck",
                "academic acid decision skin academic vanish olympic evoke gesture rumor unfair scroll grasp very steady include smell diploma package guest greatest"
                " firm humidity trial width priest class large photo sniff survive machine usher stick capacity heat improve predator float iris jacket soldier apart"
                " excuse garden cleanup realize permit dough script veteran crazy theater rival secret drink kernel lips pants",
                "academic acid decision snake acid vegan darkness bucket benefit therapy valuable impulse canyon swing distance vampire round losing twin medal treat"
                " amount fiction hush remind faint distance custody device believe campus guest preach mule exhaust regular short phrase column rescue steady float"
                " mixture testify taught fiction usher snake museum detailed agree intend inherit likely typical blimp symbolic prayer course"
            ]
        )
    }
    assert recover( details_bip39.groups['one'][1][:] + details_bip39.groups['fren'][1][:3] ) == bip39seed

    [(eth,btc)] = details_bip39.accounts
    assert eth.address == "0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E"
    assert btc.address == "bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2"

    #
    # Now, get the exact same derived accounts, but by passing the BIP-39 Seed Entropy (not the generated Seed!)
    #
    bip39entropy		= recover_bip39( 'zoo ' * 11 + 'wrong', as_entropy=True )
    assert codecs.encode( bip39entropy, 'hex_codec' ).decode( 'ascii' ) \
        == 'ff' * 16
    details_bip39entropy	= create(
        "bip39 recovery test", 2, dict( one = (1,1), two = (1,1), fam = (2,4), fren = (3,5) ),
        master_secret=bip39entropy,
        using_bip39=True,
    )
    assert recover( details_bip39entropy.groups['one'][1][:] + details_bip39entropy.groups['fren'][1][:3] ) == bip39entropy

    [(eth,btc)] = details_bip39entropy.accounts
    assert eth.address == "0xfc2077CA7F403cBECA41B1B0F62D91B5EA631B5E"
    assert btc.address == "bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2"

    #
    # Finally, test that the basic SLIP-39 encoding and derivation using the raw Seed Entropy is
    # different, and yields the expected well-known accounts.
    #
    details_slip39		= create(
        "bip39 recovery test -- all ones in SLIP-39", 2, groups_example, SEED_ONES,
    )
    #print( json.dumps( details_slip39.groups, indent=4 ))
    assert details_slip39.groups == {
        "one": (
            1,
            [
                "academic acid acrobat romp change injury painting safari drug browser trash fridge busy finger standard angry similar overall prune ladybug"
            ]
        ),
        "two": (
            1,
            [
                "academic acid beard romp believe impulse species holiday demand building earth warn lunar olympic clothes piece campus alpha short endless"
            ]
        ),
        "fam": (
            2,
            [
                "academic acid ceramic roster desire unwrap depend silent mountain agency fused primary clinic alpha database liberty silver advance replace medical",
                "academic acid ceramic scared column screw hawk dining invasion bumpy identify anxiety august sunlight intimate satoshi hobo traveler carbon class",
                "academic acid ceramic shadow believe revenue type class station domestic already fact desktop penalty omit actress rumor beaver forecast group",
                "academic acid ceramic sister actress mortgage random talent device clogs craft volume cargo item scramble easy grumpy wildlife wrist simple"
            ]
        ),
        "fren": (
            3,
            [
                "academic acid decision round academic academic academic academic academic academic academic academic academic academic academic academic academic ranked flame amount",
                "academic acid decision scatter biology trial escape element unfair cage wavy afraid provide blind pitch ultimate hybrid gravity formal voting",
                "academic acid decision shaft crunch glance exclude stilt grill numb smug stick obtain raisin force theater duke taught license scramble",
                "academic acid decision skin disaster mama alive nylon mansion listen cowboy suitable crisis pancake velvet aviation exhaust decent medal dominant",
                "academic acid decision snake aunt frozen flip crystal crystal observe equip maximum maiden dragon wine crazy nervous crystal profile fiction"
            ]
        )
    }

    # These are the well-known SLIP-39 0xffff...ffff Seed accounts
    [(eth,btc)] = details_slip39.accounts
    assert eth.address == "0x824b174803e688dE39aF5B3D7Cd39bE6515A19a1"
    assert btc.address == "bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl"

    # And ensure that the SLIP-39 encoding of the BIP-39 "zoo zoo ... wrong" w/ BIP-39
    # Entropy was identically to the raw SLIP-39 encoding.
    assert details_slip39.groups == details_bip39entropy.groups


def into_boolean( val, truthy=(), falsey=() ):
    """Check if the provided numeric or str val content is truthy or falsey; additional tuples of
    truthy/falsey lowercase values may be provided.  The empty/whitespace string is Falsey."""
    if isinstance( val, (int,float,bool)):
        return bool( val )
    assert isinstance( val, str )
    if val.strip().lower() in ( 't', 'true', 'y', 'yes' ) + truthy:
        return True
    elif val.strip().lower() in ( 'f', 'false', 'n', 'no', '' ) + falsey:
        return False
    raise ValueError( val )


def test_recover_bip39_vectors():
    # Test some BIP-39 encodings that have caused issues for other platforms:
    #
    #   - https://github.com/iancoleman/bip39/issues/58

    # If passphrase is None, signals BIP-39 recover as_entropy, and account generation using_bip39
    # bip39_tests		= [
    #     ['zoo ' * 11 + 'wrong', "", True, (
    #         None, 'bech32', 'bc1qk0a9hr7wjfxeenz9nwenw9flhq0tmsf6vsgnn2')],
    #     ['zoo ' * 11 + 'wrong', "", None, (
    #         None, 'bech32', 'bc1q9yscq3l2yfxlvnlk3cszpqefparrv7tk24u6pl')],
    #     [ 'fruit wave dwarf banana earth journey tattoo true farm silk olive fence', 'banana', True, (
    #         None, 'legacy', '17rxURoF96VhmkcEGCj5LNQkmN9HVhWb7F')]
    # ]
    with open( os.path.join( os.path.splitext( __file__ )[0] + '.csv' )) as bip32_csv:
        bip39_tests		= list( csv.DictReader( bip32_csv, quotechar='"', delimiter=',', quoting=csv.QUOTE_ALL, skipinitialspace=True ))

    for i,t in enumerate( bip39_tests ):
        log.info( f"Testing: {t!r}" )
        # Decode the entropy; either hex or BIP-39 mnemonics support
        entropy			= t.get( 'entropy' )
        if all( c in '0123456789abcdef' for c in entropy.lower() ):
            master_secret	= codecs.decode( entropy, 'hex_codec' )  # Hex entropy allowed
        else:
            passphrase		= ( t.get( 'passphrase', '' )).strip()
            if into_boolean( t.get( 'using_bip39', False )):
                # When using BIP-39, we obtain the 512-bit seed from the 128/256-bit BIP-39 mnemonic
                # entropy + passphrase, and use that to derive the wallets.
                master_secret	= recover_bip39( entropy, passphrase=passphrase )
            else:
                # For SLIP-39 wallets, we recover the 128/256-bit entropy, which is used directly as the
                # seed to derive the wallets.  The passphrase would be used to secure the SLIP-39
                # mnemonics.
                assert not passphrase, \
                    "row {i+1}: passphrase unsupported unless using_bip39"
                master_secret	= recover_bip39( entropy, passphrase=None, as_entropy=True )

        master_secret_hex	= codecs.encode( master_secret, 'hex_codec' ).decode( 'ascii' )
        log.debug( f"Entropy {entropy!r} ==> {master_secret_hex!r}" )

        # Decode the desired address; 1... and xpub... are legacy, 3... and ypub... are segwit and
        # bc1... and zpub are bech32.
        address			= t.get( 'address' )
        format			= {
            '1': 'legacy',
            'x': 'legacy',
            '3': 'segwit',
            'y': 'segwit',
            'b': 'bech32',
            'z': 'bech32',
        }[address[0]]
        acct			= account(
            master_secret	= master_secret,
            crypto		= "BTC",
            path		= t.get( 'path' ),
            format		= format
        )
        # Finally, generate the account's address and xpubkey, and see that one of them match the
        # address in the test case.  Only the address/xpubkey is in the test cases, but lets get the
        # compressed public key as well for comparision w/ the xpubkey derived account...
        addresses		= [ acct.address, acct.pubkey, acct.xpubkey ]
        assert address in addresses, \
            f"row {i+1}: BTC account {address} not in {addresses!r} for entropy {entropy} ==> {master_secret}"
        log.info( f"HD wallet path {acct.path:36}: {commas( addresses )}" )

        # OK, we recovered the same address/xpubkey as the test case did.  Now, if the path ends in
        # at least one layer of non-hardened path, we should be able to:
        #
        #     - remove one (or more) path segments, back to a hardened path component
        #     - generate the xpubkey for the same master_secret, but at that hardened path
        #     - recover a new account using that xpubkey
        #     - generate the same sub-account from it, using the remainder of the path.
        #     - confirm that the address generated by both the original (at full path) and the new
        #       account (from its xpubkey at hardened path + the remaining path) are identical (xpub
        #       will be different, but will produce the same address
        log.info( f"Testing recovery from xpub, generating sub-addresses for {acct.path}" )
        hard,soft		= path_hardened( acct.path )
        log.debug( f"BIP-44 HD path hard: {hard:14}, soft: {soft:8}" )

        acct_hard		= account(
            master_secret	= master_secret,
            crypto		= "BTC",
            path		= hard,
            format		= format
        )
        xpubkey			= acct_hard.xpubkey

        # Now recover from that xpubkey, and use the soft remainder of the path; deduces the addresses format
        acct_xpub		= account(
            master_secret	= xpubkey,
            crypto		= 'BTC',
            path		= soft,
            format		= format,
        )

        # Finally, generate the xpub-derived account's address or xpubkey, and see that one of them
        # match the address in the test case.  NOTE: the xpubkey will differ, but the address will
        # match.  This is because in the presence of the full secret key, both the public and the
        # private SECP256k1 curve points are maintained as the HD wallet path is parsed; when only
        # the public key is available, only it is modified; the secret key information is
        # unavailable, and therefore not modified as each path component is parsed.
        addresses_xpub		= [ acct_xpub.address, acct_xpub.pubkey, acct_xpub.xpubkey ]
        log.info( f"BIP-44 HD path hard: {hard:14}, soft: {soft:8}: {commas( addresses_xpub )}" )

        assert acct_xpub.address in addresses, \
            f"row {i+1}: BTC account {acct_xpub.address} not in original account's {addresses!r} for xpub-derived account"


def test_util():
    assert commas( range(10) ) == '0-9'
    assert commas( [1,2,3,5,6,7] ) == '1-3, 5-7'
    assert commas( [1,2,3,5,6,7], final='and' ) == '1-3 and 5-7'
    assert commas( [1,2,3,5,6,7,9] ) == '1-3, 5-7, 9'
    assert commas( [1,2,3,5,6,7,9], final='and' ) == '1-3, 5-7 and 9'
    assert commas( [1,3,5], final='and' ) == '1, 3 and 5'
    assert commas( [1,2,5], final='or' ) == '1, 2 or 5'

    assert round_onto( -.1, [-5, -1, 0, 1, 5], keep_sign=False ) == 0
    assert round_onto( -.9, [-5, -1, 0, 1, 5], keep_sign=False ) == -1
    assert round_onto( -10, [0, -1, -5, 5, 1], keep_sign=False ) == -5
    assert round_onto( 100, [-5, -1, 0, 1, 5], keep_sign=False ) == +5
    assert round_onto( 100, [0, -1, -5, 5, 1], keep_sign=False ) == +5
    assert round_onto( 2.9, [-5, -1, 0, 1, 5], keep_sign=False ) == +1
    assert round_onto( 3.1, [-5, -1, 0, 1, 5], keep_sign=False ) == +5
    assert round_onto( .01, [-5, -1, 0, 1, 5], keep_sign=False ) == 0
    assert round_onto( -.1, [-5, -1, 0, 1, 5], keep_sign=False ) == 0
    assert round_onto( -.1, [-5, -1, 0, 1, 5], keep_sign=True  ) == -1
    assert round_onto( -.1, [-5, -1,    1, 5], keep_sign=True  ) == -1
    assert round_onto( -.1, [-5,        1, 5], keep_sign=True  ) == -5
    assert round_onto( +.1, [-5, -1,    1, 5], keep_sign=True  ) == +1
    assert round_onto( +.1, [-5, -1,       5], keep_sign=True  ) == +5
    assert round_onto( None, [-5, -1, None, 5], keep_sign=True  ) is None


def test_dft_smoke():
    """Test some basic assumptions on DFTs"""
    print()
    print( "Real-valued samples, recovered from inverse DFT:" )
    x				= [ 2, 3, 5, 7, 11 ]
    print( "vals:  " + ' '.join( f"{f:11.2f}" for f in x ))
    y				= fft( x )
    print( "dft:   " + ' '.join( f"{f:11.2f}" for f in y ))
    z				= ifft( y )
    print( "idft:  " + ' '.join( f"{f:11.2f}" for f in z ))
    zz				= ifft( y )
    assert z == zz  # ensure any memoizing of DFT factors works
    print( "recov.:" + ' '.join( f"{f.real:11.2f}" for f in z ))

    # Ensure a DFT on a power-of-2 length sequence is the same as the pure FFT
    assert dft( x[:4] ) == pytest.approx( pfft( x[:4] ))

    # Lets determine how dft organizes its output buckets.  Lets find the bucket contain any DC
    # offset, by supplying a 0Hz signal w/ a large DC offset.

    # What about a single full real-valued waveform (complex component is always 0j)?
    print()
    print( "Real-valued signal, 1-4 cycles in 8 samples:" )
    dc				= [1.0] * 8
    print( "DC:     " + ' '.join( f"{f:11.2f}" for f in dc ))
    dft_dc			= fft( dc )
    print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_dc ))
    print( "  DFT_r:" + ' '.join( f"{f:11.2f}" for f in dft_on_real( dft_dc )))
    assert dft_dc[0] == pytest.approx( 8+0j )
    #1Hz:           1.000          0.707          0.000         -0.707         -1.000         -0.707         -0.000          0.707
    #1Hz_d: -0.000+0.000j   4.000-0.000j  -0.000-0.000j   0.000-0.000j   0.000-0.000j   0.000-0.000j   0.000-0.000j   4.000+0.000j
    oneHz			= [math.sin(+math.pi/2+math.pi*2*i/8) for i in range( 8 )]
    print( "+1Hz:   " + ' '.join( f"{f:11.2f}" for f in oneHz ))
    dft_oneHz			= fft( oneHz )
    print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_oneHz ))
    mags_oneHz			= dft_on_real( dft_oneHz )
    print( "  DFT_r:" + ' '.join( f"{f:11.2f}" for f in mags_oneHz ))
    #2Hz:           1.000          0.000         -1.000         -0.000          1.000          0.000         -1.000         -0.000
    #2Hz_d: -0.000+0.000j  -0.000-0.000j   4.000-0.000j  -0.000-0.000j   0.000-0.000j   0.000-0.000j   4.000+0.000j   0.000-0.000j
    twoHz			= [math.sin(+math.pi/2+math.pi*2*i/4) for i in range( 8 )]
    print( "+2Hz:   " + ' '.join( f"{f:11.2f}" for f in twoHz ))
    dft_twoHz			= fft( twoHz )
    print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_twoHz ))
    print( "  DFT_r:" + ' '.join( f"{f:11.2f}" for f in dft_on_real( dft_twoHz )))
    #4Hz:           1.000         -1.000          1.000         -1.000          1.000         -1.000          1.000         -1.000
    #4Hz_d:  0.000+0.000j   0.000-0.000j   0.000-0.000j   0.000-0.000j   8.000+0.000j  -0.000+0.000j   0.000-0.000j  -0.000-0.000j
    forHz			= [math.sin(+math.pi/2+math.pi*2*i/2) for i in range( 8 )]
    print( "+4Hz:   " + ' '.join( f"{f:11.2f}" for f in forHz ))
    dft_forHz			= fft( forHz )
    print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_forHz ))
    print( "  DFT_r:" + ' '.join( f"{f:11.2f}" for f in dft_on_real( dft_forHz )))
    # Same, but just shifted -PI/2, instead of +PI/2: reverses the highest frequency bucket.
    #4Hz:          -1.000          1.000         -1.000          1.000         -1.000          1.000         -1.000          1.000
    #4Hz_d:  0.000+0.000j   0.000+0.000j  -0.000+0.000j  -0.000+0.000j  -8.000-0.000j   0.000-0.000j  -0.000+0.000j   0.000+0.000j
    forHz			= [math.sin(-math.pi/2-math.pi*2*i/2) for i in range( 8 )]
    print( "-4Hz:   " + ' '.join( f"{f:11.2f}" for f in forHz ))
    dft_forHz			= fft( forHz )
    print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_forHz ))
    print( "  DFT_r:" + ' '.join( f"{f:11.2f}" for f in dft_on_real( dft_forHz )))

    # The rms(sig[:N]) energy in a signal can be obtained either from the original signal, or the rms(FFT)/sqrt(N)
    assert avg( mags_oneHz ) == pytest.approx( 0.533, abs=1e-3 )
    assert rms( oneHz ) == pytest.approx( 0.707, abs=1e-3 )
    assert rms( dft_oneHz ) / math.sqrt( len( dft_oneHz )) == pytest.approx( 0.707, abs=1e-3 )

    # So, the frequency buckets are symmetrical, from DC, 1B/N up to (N/2)B/N (which is also
    # -(N/2)B/N), and then back down to -1B/N.  We do complex signals, so we can see signals of the
    # same frequency on +'ve and -'ve side of DC.  Note that -4/8 and +4/8 are indistinguishable --
    # both rotate the complex signal by 1/2 on each steop, so the "direction" of rotation in complex
    # space is not known.  This is why the signal must be filtered; any frequency components at or
    # above B/2 will simply be mis-interpreted as artifacts in other lower frequency buckets.
    N				= 8
    print()
    print( f"Complex signal, 1-4 cycles in {N} samples:" )
    for rot_i,rot in enumerate((0, 1, 2, 3, -4, -3, -2, -1)):  # buckets, in ascending index order
        sig			= [
            # unit-magnitude complex samples, rotated through 2Pi 'rot' times, in N steps
            cmath.rect(
                1, math.pi*2*rot/N*i
            )
            for i in range( N )
        ]
        print( f"{rot:2} cyc.:" + ' '.join( f"{f:11.2f}" for f in sig ))
        dft_sig			= fft( sig )
        print( "  DFT:  " + ' '.join( f"{f:11.2f}" for f in dft_sig ))
        print( "   ABS: " + ' '.join( f"{abs(f):11.2f}" for f in dft_sig ))
        assert dft_sig[rot_i] == pytest.approx( 8+0j )


SEED_HIGH			= bytes([
    255,      1,
])
SEED_MID			= bytes([
    255,    128,      1,    128,
])
SEED_LOW			= bytes([
    255, 128+90,    128, 128-90,   1, 128-90,    128, 128+90,
])
SEED_SLOW			= bytes([
    255, 128+117, 128+90,  128+49,    128, 128-49,  128-90,  128-117,   1, 128-117, 128-90,  128-49,  128, 128+49, 128+90, 128+117
])


def test_signal_draw():
    # Detect a signal, and draw it at different rates.  First, we need to discover how to "in-fill"
    # a DFT so that the recovered signal is produced at a higher sample rate.  This allows us to
    # analyze a signal in 8-bit values, but produce a recovered signal over 2x 4-bit hex nibbles, or
    # 8x 1-bit bits.
    assert ''.join( signal_draw( s )           for s in range( -128, 128 )) \
        == ',,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,.....................................____________________________________~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'""""""""""""""""""""""""""""""""""""'
    assert ''.join( signal_draw( s, pos=True ) for s in range( -128, 128 )) \
        == '                                                                                                                                ,,,,,,,,,,,,,,,,,,,..................__________________~~~~~~~~~~~~~~~~~~~‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'""""""""""""""""""'
    assert ''.join( signal_draw( s, neg=True ) for s in range( -128, 128 )) \
        == ',,,,,,,,,,,,,,,,,,..................__________________~~~~~~~~~~~~~~~~~~~‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'\'"""""""""""""""""""                                                                                                                                '

    for entropy in (
            SEED_HIGH,     SEED_MID,     SEED_LOW,
            SEED_HIGH * 2, SEED_MID * 2, SEED_LOW * 2,
            SEED_HIGH * 3, SEED_MID * 3, SEED_LOW * 3,
            SEED_HIGH * 4, SEED_MID * 4, SEED_LOW * 4,
    ):
        print()
        print( f"Signal into DFT for {len( entropy )} bytes entropy:" )
        entropy_hex		= codecs.encode( entropy, 'hex_codec' ).decode( 'ascii' )
        entropy_bin		= ''.join( f"{int(h,16):0>4b}" for h in entropy_hex )
        symbols			= len( entropy )
        stride			= 8
        offset			= 0
        dfts			= entropy_bin_dfts( entropy_bin, offset, symbols, stride )

        print( f"Signal from DFT w/  {len( dfts )} bins:" )
        sigR			= signal_recover_real( dfts, integer=True )
        print( f"{entropy_bin}" )
        print( ''.join( stride * signal_draw( s ) for s in sigR ))

        print( "Signal from DFT x 2 deduced (for hex output):" )
        sigR2			= signal_recover_real( dfts, scale=2, integer=True )
        print( f"{entropy_hex}" )
        print( ''.join( signal_draw( s ) * 2 for s in sigR ) + " (low-frequency, expanded x2)" )
        print( ''.join( signal_draw( s ) for s in sigR2 ) + " (scaled x2)" )
        print( ''.join( signal_draw( s, neg=False ) for s in sigR2 ))
        print( ''.join( signal_draw( s, neg=True ) for s in sigR2 ))

        print( "Signal from DFT x 8 deduced (for bin output):" )
        sigR8			= signal_recover_real( dfts, scale=8, integer=True )
        print( f"{entropy_bin}" )
        sigRx8_1		= ''.join( signal_draw( s ) * 8 for s in sigR ) + " (low-frequency, expanded x8)"
        print( sigRx8_1 )
        sigR8_1			= ''.join( signal_draw( s ) for s in sigR8 ) + " (scaled x8)"
        print( sigR8_1 )
        sigR8_pos		= ''.join( signal_draw( s, neg=False ) for s in sigR8 )
        print( sigR8_pos )
        sigR8_neg		= ''.join( signal_draw( s, neg=True ) for s in sigR8 )
        print( sigR8_neg )

    # Validate the very last set of renderings
    assert sigRx8_1 \
        == '""""""""\'\'\'\'\'\'\'\'~~~~~~~~........,,,,,,,,........~~~~~~~~\'\'\'\'\'\'\'\'""""""""\'\'\'\'\'\'\'\'~~~~~~~~........,,,,,,,,........~~~~~~~~\'\'\'\'\'\'\'\'""""""""\'\'\'\'\'\'\'\'~~~~~~~~........,,,,,,,,........~~~~~~~~\'\'\'\'\'\'\'\'""""""""\'\'\'\'\'\'\'\'~~~~~~~~........,,,,,,,,........~~~~~~~~\'\'\'\'\'\'\'\' (low-frequency, expanded x8)'
    assert sigR8_1 \
        == '""""""""\'\'\'\'‾‾‾~~~___....,,,,,,,,,,,,,,,....___~~~‾‾‾\'\'\'\'"""""""""""""""\'\'\'\'‾‾‾~~~___....,,,,,,,,,,,,,,,....___~~~‾‾‾\'\'\'\'"""""""""""""""\'\'\'\'‾‾‾~~~___....,,,,,,,,,,,,,,,....___~~~‾‾‾\'\'\'\'"""""""""""""""\'\'\'\'‾‾‾~~~___....,,,,,,,,,,,,,,,....___~~~‾‾‾\'\'\'\'""""""" (scaled x8)'
    assert f"{sigR8_pos}\n{sigR8_neg}" \
        == '''\
""""""''‾‾~~__.,,                               ,,.__~~‾‾''"""""""""""''‾‾~~__.,,                               ,,.__~~‾‾''"""""""""""''‾‾~~__.,,                               ,,.__~~‾‾''"""""""""""''‾‾~~__.,,                               ,,.__~~‾‾''"""""
                 "''‾~~__..,,,,,,,,,,,..__~~‾''"                                 "''‾~~__..,,,,,,,,,,,..__~~‾''"                                 "''‾~~__..,,,,,,,,,,,..__~~‾''"                                 "''‾~~__..,,,,,,,,,,,..__~~‾''"                \
'''


def test_denoise_mags():
    """See how high we can bring up the noise level before we can no longer detect the signal. """
    print()
    symbols			= 32
    stride			= 8
    threshold			= 200/100

    # Test thru some percentage signal to noise; too high, and we'll overflow our symbols
    noise_pct			= 35

    snr_dB_strides		= {}
    for npct in range( noise_pct ):
        print()
        print( f"For {npct:2d}% noise:" )
        seed_noisy		= [ 128 + noise( 128 * npct/100 ) for _ in range( symbols ) ]  # median 1% noise floor
        for i in range( len( seed_noisy )):
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 16) ) * 16   # max frequency bin
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 14) ) * 14
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 12) ) * 12
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 10) ) * 10
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 8) ) * 8
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 6) ) * 6
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 5) ) * 5
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 4) ) * 4
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 3) ) * 3
            seed_noisy[i]      += math.cos( math.pi * 2 * i / (32 / 2) ) * 2

        SEED_NOISY		= bytes( map( int, map( round, seed_noisy )))
        entropy_hex		= codecs.encode( SEED_NOISY, 'hex_codec' ).decode( 'ascii' )
        entropy_bin		= ''.join( f"{int(h,16):0>4b}" for h in entropy_hex )

        dfts			= entropy_bin_dfts( entropy_bin, 0, symbols=symbols, stride=stride, cancel_dc=True )
        dc			= dfts[0]
        print( "dfts: " + ' '.join( f"{d:{stride*2}.1f}" for d in dfts ))
        nrms, mags		= dft_to_rms_mags( dfts )
        print( f"mags: {' '.join( f'{m:{stride*2}.1f}' for m in mags )}: {sum(mags):7.2f} sum, {avg(mags):7.2f} avg, {nrms:7.2f} RMS; dc: {dc:11.1f} == {abs(dc):7.2f} abs" )
        target, snrs		= denoise_mags( mags, 200/100 )
        snrd			= dict( snrs )  # i: snr
        print( f"snrs: {' '.join( f'{snrd[i]:{stride*2}.1f}' if i in snrd else (' ' * stride*2) for i in range( len( mags )))}: {target=:7.1f}" )
        signal			= signal_entropy( SEED_NOISY, stride=stride, symbols=symbols, threshold=threshold )
        print( f"{signal}" )

        snr_dB_strides[npct]	= (signal.dB, signal.stride)

    for npct,(dB,stride) in snr_dB_strides.items():
        print( f"{npct:02d}%: {stride:2d} bits/symbol, at {dB:7.2f}dB" )

    # We should find the signal at all these noise levels
    assert all( dBS[1] == 8 for dBS in snr_dB_strides.values() )


def test_signal_entropy():
    # See if we can detect patterns of bits in various frequency bins.  With 8x 8-bit real-valued
    # samples, we get DC + 4 frequency bins, with the highest frequency bin representing changes in
    # samples at the peak Nyquist rate; every 2 symbols.  The lowest
    #
    #               256
    #       128+90      128+90
    #   128                      128
    #                                128-90      128+90
    #                                         0
    #
    # We can estimate what perfect noise looks like across all energy bins, and compare our signal
    # against that baseline.  So, our signal has some noise added to it.
    print()
    SEED_SINE			= bytes([
        # 255,    255,    255,    255,
        # 255,    255,    255,    255,
        # 128,    128,    128,    128,
        # 128,    128,    128,    128,
        #   0,      0,      0,      0,
        #   0,      0,      0,      0,
        129,      127,    128,    126,  # 0
        128+20,   128, 128-21,    128,  # 32
        128+21,   128, 128-19,    128,  # 64
        131,      129,    126,    127,  # 96
        127,      129,    126,    129,  # 128
        128+90,   255, 128+90,    128,  # 160
        128-90,     0, 128-90,    128,  # 192
        127,      130,    128,    129,  # 224
    ])

    # Find the highest amplitude signal over 8-symbol chunks; should be the later, lower-frequency
    signal			= signal_entropy( SEED_SINE[:16],  8, 8, threshold=300/100 )
    print( f"{signal}" )
    signal			= signal_entropy( SEED_SINE[-16:], 8, 8, threshold=300/100 )
    print( f"{signal}" )
    signal			= signal_entropy( SEED_SINE,       8, 8, threshold=300/100 )
    print( f"{signal}" )
    assert signal.offset == 160
    assert signal.dB == pytest.approx( 8.7, abs=1e-1 )

    SEED_FF00			= codecs.decode( '7f81' + 'ff00' * 7, 'hex_codec' )
    signal			= signal_entropy( SEED_FF00, 8, 8, threshold=300/100 )
    print( f"SEED_FF00: {signal}" )
    assert signal.offset == 16
    assert signal.dB == pytest.approx( 8.7, abs=1e-1 )
    signal			= signal_entropy( SEED_FF00, 4, 16, threshold=300/100 )
    print( f"SEED_FF00: {signal}" )
    assert signal.offset == 16
    assert signal.dB == pytest.approx( 17.5, abs=1e-1 )

    SEED_F0F0			= codecs.decode( '7f818081' + 'f0f0' * 6, 'hex_codec' )
    signal			= signal_entropy( SEED_F0F0, 8, 8, threshold=300/100 )
    print( f"SEED_F0F0: {signal}" )
    assert signal.offset == 32
    assert signal.dB == pytest.approx( 8.9, abs=1e-1 )
    signal			= signal_entropy( SEED_F0F0, 4, 16, threshold=300/100 )
    print( f"SEED_F0F0: {signal}" )
    assert signal.offset == 32
    assert signal.dB == pytest.approx( 17.0, abs=1e-1 )

    SEED_RAMP			= codecs.decode( '000102030405060708090A0B0C0D0E0F', 'hex_codec' )
    signal			= signal_entropy( SEED_RAMP, 8, threshold=300/100 )
    print( f"SEED_RAMP: {signal}" )
    assert signal.dB == pytest.approx( 15.6, abs=1e-1 )
    signal			= signal_entropy( SEED_RAMP, 4, threshold=250/100 )
    print( f"SEED_RAMP: {signal}" )
    assert "DC offset and every 2 symbols" in signal.details
    assert signal.dB == pytest.approx( 3.7, abs=1e-1 )

    signal			= signal_entropy( SEED_XMAS, threshold=300/100 )
    print( f"SEED_XMAS: {signal}" )
    assert signal.dB == pytest.approx( -2.2, abs=1e-1 )

    signal			= signal_entropy( SEED_XMAS, threshold=300/100, overlap=True )
    print( f"SEED_XMAS: {signal}" )
    assert signal.dB == pytest.approx( -1.5, abs=1e-1 )

    analysis			= analyze_entropy( SEED_XMAS )
    print( f"SEED_XMAS: {analysis}" )
    assert analysis is None


def test_signal_lots():
    # Test that lots of weaker signals still above the threshold are included in the detected
    # Signal.dB ratio.  First, weaker signals just below the threshold, and a couple above.
    # Also confirm that we're normalizing the real-valued magnitudes correctly.
    print()
    seed_lots			= [ 128 + noise( 128 * 1/100 ) for _ in range( 32 ) ]  # median 1% noise floor
    for i in range( len( seed_lots )):
        seed_lots[i]	       += math.cos( math.pi * 2 * i / 2 ) * 16   # max frequency bin
        seed_lots[i]	       += math.cos( math.pi * 2 * i / 8 ) * 4
        seed_lots[i]	       += math.cos( math.pi * 2 * i / (10+2/3) ) * 16  # 32/3 -- exactly the 3rd harmonic
        seed_lots[i]	       += math.cos( math.pi * 2 * i / 32 ) * 16   # low frequency bin

    SEED_LOTS			= bytes( map( int, map( round, seed_lots )))
    stride			= 8
    # Make sure we're computing the complex DFTs to real bins correctly; the same energy in the DC /
    # highest frequency bin (single), and the energy split between +'ve/-'ve frequency complex bins
    # summed into one real magnitude bin, should yield the same magnitudes.
    dfts			= fft( seed_lots )  # DC offset not canceled
    print( f"dfts: {' '.join( f'{b:{stride*2}.1f}' for b in dfts )}" )
    dc				= dfts[0]
    mags			= dft_on_real( dfts )
    print( f"mags: {' '.join( f'{m:{stride*2}.1f}' for m in mags )}: {sum(mags):7.2f} sum, {avg(mags):7.2f} avg, dc: {dc:11.1f} == {abs(dc):7.2f} abs" )
    assert mags[1] == pytest.approx( mags[-1], rel=10/100 )

    signal			= signal_entropy( SEED_LOTS, stride, threshold=300/100, harmonics_max=None )
    print( f"SEED_LOTS: {signal}" )
    assert signal.dB == pytest.approx( +14.8, rel=20/100 )
    # assert "every 32, 10+2/3 and 2 symbols" in signal.details  # order is non-deterministic

    # Now add a strong signal exactly between 2 bins.  The energy of all three signals above
    # threshold should accrue to the result'st SNR dB.  The one that is split across 2 bins should
    # sum to about the same energy as the one concentrated in one bin.
    seed_midl			= [128.] * 32
    midl			= 32/9.5                # harmonics at every 3+5/9 and 3+1/5 symbols
    othr			= 32/5			# and every 6+2/5 symbols
    for i in range( len( seed_lots )):
        seed_midl[i]	       += math.cos( math.pi * 2 * i / midl ) * 60
        seed_midl[i]	       += math.cos( math.pi * 2 * i / othr ) * 40

    SEED_MIDL			= bytes( map( int, map( round, seed_midl )))
    signal			= signal_entropy( SEED_MIDL, stride, threshold=300/100, harmonics_max=None )
    print( f"SEED_MIDL: {signal}" )
    assert signal.dB == pytest.approx( +4.1, abs=1e-1 )
    assert "every 6+2/5, 3+5/9 and 3+1/5 symbols" in signal.details


def test_shannon_entropy():
    shannon			= shannon_entropy( SEED_ONES )
    shannon			= shannon_entropy( SEED_ONES, overlap=False )
    shannon			= shannon_entropy( SEED_ONES, stride=4 )
    shannon			= shannon_entropy( SEED_ONES, stride=4, overlap=False )

    shannon			= shannon_entropy( SEED_ONES+SEED_ZERO )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_ZERO, overlap=False )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_ZERO, stride=4 )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_ZERO, stride=4, overlap=False )
    print( f"{shannon}" )

    shannon			= shannon_entropy( SEED_ONES+SEED_XMAS )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_XMAS, overlap=False )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_XMAS, stride=4 )
    print( f"{shannon}" )
    shannon			= shannon_entropy( SEED_ONES+SEED_XMAS, stride=4, overlap=False )
    print( f"{shannon}" )

    shannon			= shannon_entropy( SEED_XMAS )
    shannon			= shannon_entropy( SEED_XMAS, stride=6 )
    shannon			= shannon_entropy( SEED_XMAS, stride=6, overlap=False )
    shannon			= shannon_entropy( SEED_XMAS, stride=4 )
    assert shannon.dB == pytest.approx( -7.95, abs=1e-1 )
    shannon			= shannon_entropy( SEED_XMAS )
    assert shannon.dB == pytest.approx( -40.0, abs=1e-1 )
    # Now, add some duplicates, reducing the entropy, 'til we fail the Shannon entropy test.
    shannon			= shannon_entropy( SEED_XMAS[:-3]+SEED_XMAS[:3] )
    print( f"{shannon}" )
    assert shannon.dB == pytest.approx( -3.4, abs=1e-1 )
    shannon			= shannon_entropy( SEED_XMAS[:-4]+SEED_XMAS[:4] )
    print( f"{shannon}" )
    assert shannon.dB == pytest.approx( -0.93, abs=1e-1 )
    shannon			= shannon_entropy( SEED_XMAS[:-5]+SEED_XMAS[:5] )
    print( f"{shannon}" )
    assert shannon.dB == pytest.approx( +0.98, abs=1e-1 )


def kwargs_shannon_limits( kwargs ):
    return compute_entropy_limits( shannon_entropy, **kwargs )


def kwargs_signal_limits( kwargs ):
    return compute_entropy_limits( signal_entropy, **kwargs )


def compute_entropy_limits( compute_entropy, bits, overlap, stride, threshold, setpoint, cycles, checks, symbols=None ):
    avg_over		= (4096, 8192, 16384)
    rejects		= deque( maxlen=16384 )
    rejected		= setpoint
    try:
        for i in range( cycles ):
            entropy	= os.urandom( bits // 8 )
            threshold  *= 1 + ( rejected - setpoint ) / ( 10**(1+math.log(i+1,10)/6) )  # / (10...1000) as i increases from 0 to 100000
            if symbols is None:
                signal	= compute_entropy( entropy, overlap=overlap, stride=stride,
                                           threshold=threshold, show_details=False )
            else:
                signal	= compute_entropy( entropy, overlap=overlap, stride=stride, symbols=symbols,
                                           threshold=threshold, show_details=False )
            reject	= signal.dB >= 0
            rejects.append( reject )
            rejects_lst	= list( rejects )
            rejected	= sum( avg(rejects_lst[-n:]) for n in avg_over ) / len( avg_over )
            if reject or i % checks == 0:
                ( log.info if i % checks == 0 else log.debug )(
                    f" - {i:6} {threshold=:7.5f}: {100*rejected:7.3f}%; "
                    + ', '.join( f"{100*avg(rejects_lst[-n:]):7.3f}%/{n}" for n in avg_over )
                    + f": {signal}"
                )
    except Exception as exc:
        threshold	= exc
    finally:
        print( f"{compute_entropy.__name__} for {bits}-bit entropy w/ {overlap=:5}, {stride=:3}; {rejected*100:7.3f}% (latest avg), {avg(rejects)*100:7.3f}% rejects total: {threshold=}" )
        return (bits, overlap, stride, symbols, threshold)


def test_shannon_limits( detailed=False ):
    """Compute the threshold at which ~99.9% of good random entropy are accepted, at each combination
    of stride.  Only runs full test only at high logging levels.

    """
    shannon_limits		= {}
    strengths			= (128, 256, 512) + (160, 192, 224)  # SLIP-39 + BIP-39
    strides			= (3, 4, 5, 6, 7, 8)
    overlapping			= (False, True)
    cycles			= 150001
    checks			= 10000
    if not detailed:
        strengths		= strengths[-1:]
        strides			= strides[-2:]
        #overlapping		= overlapping[:1]
        cycles			= 1001
        checks			= 100

    threshold			= 10/100
    setpoint			= 0.15/100
    poolkwargs			= []
    for bits in strengths:
        for overlap in overlapping:  # noqa: E111
            for stride in strides:
                poolkwargs.append( dict(
                    bits	= bits,
                    overlap	= overlap,
                    stride	= stride,
                    threshold	= threshold,
                    setpoint	= setpoint,
                    cycles	= cycles,
                    checks	= checks,
                ))
    with multiprocessing.Pool(32) as pool:
        for (bits, overlap, stride, symbols, threshold) in pool.map( kwargs_shannon_limits, poolkwargs ):
            if isinstance( threshold, Exception ):
                print( f"Shannon Limits: Failed for {bits=:3}, {overlap=:5}, {symbols=:3}: {threshold}" )
            else:
                shannon_limits.setdefault(
                    overlap, {} ).setdefault(
                        bits, {} )[stride] = threshold
    print( f"Shannon limits: {json.dumps( shannon_limits, indent=4, default=str, sort_keys=True )}" )


def test_signal_limits( detailed=False ):
    """Compute the threshold at which ~99.9% of good random entropy are accepted, at each combination
    of stride.  Only runs full test only at high logging levels.

    """
    signal_limits		= {}
    strengths			= (128, 256, 512) + (160, 192, 224)  # SLIP-39 + BIP-39
    strides			= (3, 4, 5, 6, 7, 8)
    overlapping			= (False, True)
    cycles			= 150001
    checks			= 10000
    if not detailed:
        strengths		= strengths[:1]
        strides			= strides[:-2]
        #overlapping		= overlapping[:1]
        cycles			= 1001
        checks			= 100

    threshold			= 300/100
    setpoint			= 0.15/100
    poolkwargs			= []
    for bits in strengths:
        for overlap in overlapping:  # noqa: E111
            for stride in strides:
                for symbols in range( bits // stride - 3, bits // stride + 1 ):
                    poolkwargs.append( dict(
                        bits		= bits,
                        overlap		= overlap,
                        symbols		= symbols,
                        stride		= stride,
                        threshold	= threshold,
                        setpoint	= setpoint,
                        cycles		= cycles,
                        checks		= checks,
                    ))
    with multiprocessing.Pool(32) as pool:
        for (bits, overlap, stride, symbols, threshold) in pool.map( kwargs_signal_limits, poolkwargs ):
            if isinstance( threshold, Exception ):
                print( f"Signal Limits: Failed for {bits=:3}, {overlap=:5}, {symbols=:3}: {threshold}" )
            else:
                signal_limits.setdefault(
                    overlap, {} ).setdefault(
                        bits, {} ).setdefault(
                            stride, {} )[symbols] = threshold
    print( f"Signal limits: {json.dumps( signal_limits, indent=4, default=str, sort_keys=True )}" )


def test_poor_entropy():
    """Test various entropy to determine if we see a failure of Shannon entropy or Signals.

    """
    # Some really bad entropy that perfectly match the highest frequency DFT bins in an 16-symbol
    # complex DFT (DC+5 magnitudes real).  Try to determine how to determine the frequency buckets.

    entropy			= SEED_HIGH * 8
    signal			= signal_entropy( entropy, stride=8, symbols=16, threshold=300/100, overlap=False, ignore_dc=True )
    log.info( f"Signal high frequency: {signal}" )
    assert signal.dB == pytest.approx( +19.1, abs=1e-1 )

    entropy			= SEED_MID * 4
    signal			= signal_entropy( entropy, stride=8, symbols=16, overlap=False, ignore_dc=True )
    log.info( f"Signal mid. frequency: {signal}" )
    assert signal.dB == pytest.approx( +14.7, abs=1e-1 )

    entropy			= SEED_LOW * 2
    signal			= signal_entropy( entropy, stride=8, symbols=16, overlap=False, ignore_dc=True )
    log.info( f"Signal low  frequency: {signal}" )
    assert signal.dB == pytest.approx( +14.7, abs=1e-1 )

    entropy			= SEED_SLOW
    signal			= signal_entropy( entropy, stride=8, symbols=16, overlap=False, ignore_dc=True )
    log.info( f"Signal slow  frequency: {signal}" )
    assert signal.dB == pytest.approx( +14.6, abs=1e-1 )

    # Some bad "random" dice rolls.  These are ASCII data, so only 8-bit symbol strides, and ignore
    # DC offset.
    entropy			= "34131214324563463456112412364563".encode( 'ASCII' )
    signal			= signal_entropy( entropy, stride=8, symbols=32, overlap=False, threshold=250/100, ignore_dc=True )
    log.info( f"Signal in bad dice rolls: {signal}" )
    assert signal.dB == pytest.approx( +2.1, abs=1e-1 )
    assert "every 16 symbols" in signal.details

    analysis			= analyze_entropy( entropy, strides=8, overlap=False, ignore_dc=True )
    print( f"Analysis of bad dice rolls: {analysis}" )

    # Some bad "random" entropy in a base-64 phrase; see if we can pick it out
    entropy			= base64.b64decode( "The-quick-brown-fox-jumps-over-the-lazy-dog=", altchars='-_', validate=True )
    signal			= signal_entropy( entropy, stride=6, overlap=False, threshold=200/100, ignore_dc=True )
    log.info( f"Signal in bad base-64 string: {signal}" )
    assert signal.dB == pytest.approx( +0.4, abs=1e-1 )
    assert "every 5+1/4 symbols" in signal.details

    signal			= shannon_entropy( entropy, stride=6, overlap=True, threshold=10/100 )
    log.info( f"Shannon in bad base-64 string: {signal}" )
    assert signal.dB == pytest.approx( +5.9, abs=1e-1 )

    signals, shannons		= scan_entropy( entropy, shannon_threshold=10/100 )
    #assert len( signals ) == 3
    assert len( shannons ) == 3

    analysis			= analyze_entropy( entropy )
    print( f"Analysis of base-64 phrase: {analysis}" )
    assert analysis and "Shannon entropy reduced at offset 5 in 41x 6-bit symbols" in analysis


def test_good_entropy():
    entropy			= SEED_XMAS
    signal			= signal_entropy( entropy )
    log.info( f"Signal XMAS frequency: {signal}" )
    assert signal.dB == pytest.approx( -4.3, abs=1e-1 )

    analysis			= analyze_entropy( entropy )
    print( f"Analysis of XMAS entropy: {analysis}" )
    assert analysis is None

    analysis			= analyze_entropy( entropy[:4]
                                                   + 2 * codecs.decode( "DeadBeef", 'hex_codec' )
                                                   + entropy[12:])
    print( f"Analysis of XMAS entropy w/ 0xDeadBeef: {analysis}" )

    analysis			= analyze_entropy( entropy[:-5] + entropy[:5] )
    print( f"Analysis of XMAS entropy w/ 5 dups: {analysis}" )
    assert analysis and "Shannon entropy reduced" in analysis

    # This test takes a while without numpy installed, due to inefficient python-only dft
    cycles			= 1000
    analysis_bad		= []
    signals_bad			= []
    shannon_bad			= []
    for i in range( cycles ):
        entropy			= secrets.token_bytes( 256 // 8 )
        analysis		= analyze_entropy( entropy )
        if analysis:
            analysis_bad.append( analysis )
            if "Signal" in analysis:
                signals_bad.append( analysis )
                print( f"{ordinal(i)} analysis shows Signal energy: {analysis}" )
            if "Shannon" in analysis:
                shannon_bad.append( analysis )

    print( f"Analyzed {cycles} random entropy and found {100*len(analysis_bad)/cycles:.1f}% bad" )
    print( f"  Signals failure found {100*len(signals_bad)/cycles:.1f}%" )
    print( f"  Shannon failure found {100*len(shannon_bad)/cycles:.1f}%" )
    assert len( analysis_bad ) / cycles < 5/100
    assert len( shannon_bad ) / cycles < 3/100
    assert len( signals_bad ) / cycles < 3/100


def test_rngs_entropy( detailed=False ):
    """Test various RNGs to observe that they exhibit similar spectral features.

    """
    def rng_secrets( n ):
        return secrets.token_bytes( n )

    def rng_os( n ):
        return os.urandom( n )

    def rng_py( n ):
        return rng_py.R.randbytes( n )
    rng_py.R			= random.Random(
        os.urandom( 32 )
    )

    def rng_hash( n ):
        """Introduce n bytes of entropy into the SHA512, and then extract stretched entropy from the low
        half of the digest, re-hashing as necessary.  Basically a poor RNG based on SHA512.

        """
        def hasher():
            while True:
                digest		= rng_hash.H.digest()
                for i in range( 256 ):
                    yield digest[i:i+1]
                rng_hash.H.update( digest )
        rng_hash.H.update( rng_os( n ))
        return b''.join( itertools.islice( hasher(), n ))
    rng_hash.H			= hashlib.sha512()

    def rng_recycle( n ):
        """Every random 10 or so bytes, recycle a byte."""
        def recycler():
            while True:
                b		= rng_os( 2 )
                if b[1] % 10 == 0:
                    yield b[:1]
                yield b[:1]
        return b''.join( itertools.islice( recycler(), n ))

    def eval_rng( rng, count, bits=256 ):
        stat			= dict(
            shannons_dBs= [],
            signals_dBs	= [],
        )
        for i in range( count ):
            entropy		= rng( bits // 8 )
            signals, shannons	= scan_entropy( entropy )
            stat['signals_dBs'].append( signals )
            stat['shannons_dBs'].append( shannons )
        return stat

    summ			= {}
    count			= 1000
    if not detailed:
        count			= 10
    for rng in ( rng_os, rng_secrets, rng_py, rng_hash, rng_recycle ):
        stat			= eval_rng( rng, count )

        summ.setdefault( rng.__name__, {} )['signals pct'] \
            = avg( list( bool( s ) for s in stat['signals_dBs'] ))
        summ.setdefault( rng.__name__, {} )['shannons pct'] \
            = avg( list( bool( s ) for s in stat['shannons_dBs'] ))
        summ.setdefault( rng.__name__, {} )['signals avg'] \
            = avg( list( s[0].dB for s in stat['signals_dBs'] if s ))
        summ.setdefault( rng.__name__, {} )['shannons avg'] \
            = avg( list( s[0].dB for s in stat['shannons_dBs'] if s ))

        for n in 'signals', 'shannons':
            for i,s in enumerate( sorted(
                    (
                        s[0] for s in stat[n + '_dBs']
                        if s
                    ),
                    reverse=True
            )[:3] ):
                print( f"{ordinal(i)} {rng.__name__}: {s}" )

    print( f"Summary (goal is ~1% of entropy reports signals/shannon failure): {json.dumps( summ, indent=4)}" )


if __name__ == "__main__":
    #import cProfile
    #cProfile.run( 'test_signal_limits( detailed=True ); test_shannon_limits( detailed=True )' )
    #cProfile.run( 'test_shannon_limits( detailed=True )' )
    #test_shannon_limits( detailed=True )
    test_signal_limits( detailed=True )




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#

from __future__		import annotations

import logging
import textwrap

from typing		import Union
from itertools		import count

from web3		import Web3		# noqa F401

#from tabulate		import tabulate
from ..tabulate		import tabulate

from ..util		import remainder_after, fraction_allocated, commas
from .ethereum		import Chain, Contract, contract_address  # noqa F401

__author__                      = "Perry Kundert"
__email__                       = "perry@dominionrnd.com"
__copyright__                   = "Copyright (c) 2022 Dominion Research & Development Corp."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

log				= logging.getLogger( 'multipayout' )


def payout_reserve( addr_frac, bits=16 ):
    """Produce recipients array elements w/ fixed fractional/proportional amounts to predefined
    addresses.  Convert proportions in the range [0,1) totaling to exactly 1.0, to a remainder
    fraction within a multiple of 2^bits, or will raise Exception.  It is not valid to specify
    a recipient with a zero proportion, and the minimum allocation is 1/2^bits.

    The incoming { addr: proportion,... } dict values will be normalized, so any numeric/fractional
    values may be used (eg. arbitrary numeric "shares").

    So long as the final remainder is within +/- 1/scale of 0, its int will be zero, and
    this function will succeed.  Otherwise, it will terminate with a non-zero remainder,
    and will raise an Exception.

    The Gas cost of shifts is 3 (FASTESTSTEP), divisions is 5 (FASTSTEP), so striving to use a scale
    multiplier that is a power of 2 and known at compile-time would be best: eg. a fixed-point
    denominator that is some factor of the bit-size of the numerator.

    Since we're producing fractions in the range (0,1], the fixed-point denominator can be the 2^N
    for an N-bit value supporting the range (0,2^N-1) -- the largest possible fraction we can represent
    will be (2^N-1)/(2^N).  Eg, for a 16-bit value:

        65535/65536 =~= 0.9999847412
            1/65536 =~= 0.0000152588

    """
    addr_frac_sorted		= sorted( addr_frac.items(), key=lambda a_p: a_p[1] )
    addresses,fractions		= zip( *addr_frac_sorted )
    fractions_total		= sum( fractions )
    rem				= None
    for addr,frac,rem in zip(
            addresses,
            fractions,
            remainder_after(
                fractions,
                scale	= 2 ** bits / fractions_total,  # If Fraction, will remain a Fraction
                total	= 2 ** bits,
            )
    ):
        assert frac > 0, \
            f"Encountered a zero proportion: {frac} for recipient {addr}"
        assert int( rem ) < 2 ** bits, \
            f"Encountered a remainder fraction: {rem} numerator greater or equal to the denominator: {2 ** bits}"
        yield addr,frac/fractions_total,rem

    assert rem is not None, \
        "At least one payee is required"
    assert int( rem ) == 0, \
        f"Total payout percentages didn't accumulate to zero remainder: {rem:7.4f} =~= {int( rem )}, for {commas( fractions, final='and' )}"


class MultiPayoutERC20( Contract ):
    """Provide access to our MultiPayoutERC20 Contract; either a pre-existing deployed Contract via
    address, or _deploy a new contract by specifying payees and optionally a list of erc20s to
    support.

        >>> from .ethereum import Chain, alchemy_url
        >>> mp = MultiPayoutERC20( Web3.WebsocketProvider( alchemy_url( Chain.Goerli )),
        ...    address		= "0x8b3D24A120BB486c2B7583601E6c0cf37c9A2C04"
        ... )
        MultiPayoutERC20 Payees:
            | Payee                                      | Share                 |   Frac. % |   Reserve |   Reserve/2^16 |   Frac.Rec. % |   Error % |
            |--------------------------------------------+-----------------------+-----------+-----------+----------------+---------------+-----------|
            | 0x7Fc431B8FC8250A992567E3D7Da20EE68C155109 | 4323/65536            |   6.59638 |     61213 |          61213 |       6.59638 |         0 |
            | 0xEeC2b464c2f50706E3364f5893c659edC9E4153A | 1507247699/4294967296 |  35.0933  |     40913 |          40913 |      35.0933  |         0 |
            | 0xE5714055437154E812d451aF86239087E0829fA8 | 2504407469/4294967296 |  58.3103  |         0 |              0 |      58.3103  |         0 |
        ERC-20s:
            | Token                                      | Symbol   |   Digits |
            |--------------------------------------------+----------+----------|
            | 0xe802376580c10fE23F027e1E19Ed9D54d4C9311e | USDT     |        6 |
            | 0xde637d4C445cA2aae8F782FFAc8d2971b93A4998 | USDC     |        6 |
            | 0xaFF4481D10270F50f203E0763e2597776068CBc5 | WEENUS   |       18 |
            | 0x1f9061B953bBa0E36BF50F21876132DcF276fC6e | ZEENUS   |        0 |

        >>> import json
        >>> print( json.dumps( mp._payees, indent=4, default=lambda f: f"{float( f * 100 ):9.5f}%" ))
        {
            "0x7Fc431B8FC8250A992567E3D7Da20EE68C155109": "  6.59637%",
            "0xEeC2b464c2f50706E3364f5893c659edC9E4153A": " 35.09335%",
            "0xE5714055437154E812d451aF86239087E0829fA8": " 58.31028%"
        }
        >>> print( json.dumps( mp._erc20s_data, indent=4 ))
        {
            "0xe802376580c10fE23F027e1E19Ed9D54d4C9311e": [
                "USDT",
                6
            ],
            "0xde637d4C445cA2aae8F782FFAc8d2971b93A4998": [
                "USDC",
                6
            ],
            "0xaFF4481D10270F50f203E0763e2597776068CBc5": [
                "WEENUS",
                18
            ],
            "0x1f9061B953bBa0E36BF50F21876132DcF276fC6e": [
                "ZEENUS",
                0
            ]
        }

    TODO: This is a quite complex test with a lot of prerequisites; perhaps move from doctest to pytest

    """
    def __init__( self, *args, address=None, payees=None, erc20s=None, gas=2000000, **kwds ):
        """Either connect to an existing Contract by address, or deploy the contract w/ the specified payees
        and (initial) ERC-20s.

        If deploying, produces the _payees_data from the original fractions, and can deduce the
        "error" between the specified fractions and the resultant payout fractions based on
        decimated fixed-point "reserve" values.  If reading from existing deployed contract,
        no error can be calculated as only "reserve" values are available.

        """
        assert bool( address ) ^ bool( payees or erc20s ), \
            "Either a Contract address, or details of desired payees and optionally ERC-20s must be supplied"

        self._payees		= payees or dict()      # { <Payee address>: <Fraction>, ... }
        self._erc20s		= erc20s or list()      # [ <ERC-20 address>: ("<symbol>",<decimals>), ... ]
        self._erc20s_data	= dict()
        self._forwarder_hash	= None			# Keccack has of the Forwarder constructor bytecode for CREATE2
        self._bits		= 16			# reserve values denominated in 1/2^bits

        super().__init__(
            *args, address=address, **kwds		# If address supplied, will invoke self._update()
        )

        if not self._address:
            # Otherwise, will deploy Contract from computed payees reserve values and erc20s list,
            # and then invoke self._update
            self._deploy( self._payees_reserve, self._erc20s, gas=gas )

        print( self )

    def __str__( self ):
        return f"""\
{self.__class__.__name__} Payees:
{textwrap.indent( self._payees_table, ' ' * 4 )}
ERC-20s:
{textwrap.indent( self._erc20s_table, ' ' * 4 )}"""

    @property
    def _erc20s_table( self ):
        return tabulate(
            list(
                (tok,sym,dig)
                for tok,(sym,dig) in self._erc20s_data.items()
            ),
            headers	= [ 'Token', 'Symbol', 'Digits' ],
            tablefmt	= 'orgtbl'
        )

    @property
    def _payees_data( self ):
        """The original proportion assigned to each address is in shares/fractions; the reserve
        remaining after each payout is in rsvs, and this actual payout percentage vs. recovered
        payout percentage error is computed later in 'error'.  This may differ from the exact
        original percentages above, by < 1/2^bits due to fixed point math.  If the payee fractions
        are recovered from an existing contract, of course these will be exact (no "error").

        """
        addrs,fracs,rsrvs	= zip( *payout_reserve( self._payees, bits=self._bits ))
        share			= list( self._payees[a] for a in addrs )
        rsrvs_int		= list( map( int, rsrvs ))
        fracs_recov		= list( fraction_allocated( rsrvs_int, scale=2**self._bits ))
        # Now, compute the "error" between the originally specified payout fractions, and the payout
        # fractions recovered after reversing the 1/2^bits reserve calculations.  The error should
        # always be < 1/2^bits.
        fracs_error		= list( fr - f for f,fr in zip( fracs, fracs_recov ))

        return dict(
            zip(
                addrs,					# The Payee address
                zip(
                    share,				# Original "Share" value specified
                    fracs,				# Fraction computed [0,1), full resolution
                    rsrvs,				# Reserve after each payee, full resolution
                    rsrvs_int,				# Reserve, denominated in 1/2^bits
                    fracs_recov,			# Fraction recovered in 1/2^bits resolution
                    fracs_error,
                )
            )
        )

    @property
    def _payees_reserve( self ):
        return list(
            (a,ri)
            for a,(s,f,r,ri,fr,fe) in self._payees_data.items()
        )

    @_payees_reserve.setter
    def _payees_reserve( self, payees_reserve ):
        """Recovers payee payout fractions from fixed-point (decimated) reserves (eg. obtained from an
        existing contract's data).  If we already have self._payees, just confirms that exactly the same
        reserves are computed!

        This provides a round-trip confirmation of our originally specified payees fractions, to
        decimated reserves and fractions, through contract deployment, and then recovery of
        payees/reserve from the contract, to recomputed decimated fractions.

        """
        addrs,reserve		= zip( *payees_reserve )
        payees			= dict(
            zip(
                addrs,
                fraction_allocated(
                    reserve,
                    scale	= 2**self._bits,
                )
            )
        )
        if self._payees:
            payees_current	= dict(
                (a,fr)
                for a,(s,f,r,ri,fr,fe) in self._payees_data.items()
            )
            if payees != payees_current:
                log.warning( f"Recovered payees fractions don't match: \n{self._payees_table}" )
        else:
            self._payees	= payees

    @property
    def _payees_table( self ):
        return tabulate(
            list(
                (a,f"{s}",f"{float(f*100):10.6f}",f"{float(r):10.6f}",ri,f"{float(fr*100):10.6f}",f"{float(fe*100):10.6f}")
                for a,(s,f,r,ri,fr,fe) in self._payees_data.items()
            ),
            headers	= [ 'Payee', 'Share', 'Frac. %', 'Reserve', f"Reserve/2^{self._bits}",'Frac.Rec. %','Error %' ],
            tablefmt	= 'orgtbl'
        )

    def _forwarder_address_precompute(
        self,
        salt: Union[int,bytes,str]
    ):
        """We have all the data required to precompute the CREATE2 ...Forwarder contract address."""
        assert self._forwarder_hash, \
            "Cannot precompute ...Forwarder addresses until MultiPayoutERC20 Contract is deployed and its address is known."
        return contract_address(
            address		= self._address,
            salt		= salt,
            creation_hash	= self._forwarder_hash
        )

    def _update( self ):
        """Query the deployed MultiPayoutERC20 Contract to get payees and the (currently) specified
        ERC-20 tokens (<symbol>, <decimals>).

        Computes the Fraction representing the portion paid to each payee, from the contract's list
        of payees, and the remainder left to subsequent payees.  The Contract enforces that the
        final payee has a 0 reserve.

        TODO: recursively descend into payees that are also MultiPayoutERC20 contracts, and deduce
        the full tree of payees, and their aggregate percentages.

        """
        self._forwarder_hash	= self.forwarder_hash()

        def payees_reserve():
            for i in count():
                payee, reserve	= self.payees( i )
                yield payee, reserve
                if not reserve:
                    break

        # This only *needs* to be done at initial _update from an existing deployed contract.
        # However, we'll always recover them here (even if we just deployed the contract), to
        # make certain our reverse Reserve to Fraction calculation is correct.  Computes
        # self._payees if absent.
        self._payees_reserve	= list( payees_reserve() )

        def erc20s_data():
            # Look up the contract interface we've imported as IERC20Metadata, and use its ABI
            # for accessing any ERC-20 tokens' symbol and decimals.
            IERC20_key		= self._abi_key( "IERC20Metadata" )
            for i in range( self.erc20s_len()):
                tok		= self.erc20s( i )
                IERC20_tok	= self._w3.eth.contract(
                    address	= tok,
                    abi		= self._compiled[IERC20_key]['abi'],
                )
                # These should be free calls (public data or view-only functions)
                sym			= IERC20_tok.functions.symbol().call()
                dig			= IERC20_tok.functions.decimals().call()
                yield tok,(sym,dig)

        self._erc20s_data	= dict( erc20s_data() )
        erc20s			= list( self._erc20s_data )
        if self._erc20s:
            assert self._erc20s == erc20s
        else:
            self._erc20s	= erc20s




#
# Python-slip39 -- Ethereum SLIP-39 Account Generation and Recovery
#
# Copyright (c) 2022, Dominion Research & Development Corp.
#
# Python-slip39 is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.  It is also available under alternative (eg. Commercial) licenses, at
# your option.  See the LICENSE file at the top of the source tree.
#
# Python-slip39 is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
from __future__          import annotations

import logging
import math

from dataclasses	import dataclass
from collections	import namedtuple, defaultdict
from typing		import Dict, Union, Optional, Sequence, List, Any, Tuple
from fractions		import Fraction
from pathlib		import Path
from datetime		import datetime, timedelta, timezone
from calendar		import monthrange

import fpdf
from ..			import tabulate

from crypto_licensing.misc import get_localzone, Duration

from ..api		import Account
from ..util		import commas, is_listlike, is_mapping
from ..defaults		import (
    INVOICE_CURRENCY, INVOICE_ROWS, INVOICE_STRFTIME, INVOICE_DUE, INVOICE_DESCRIPTION_MAX,
    INVOICE_FORMAT, MM_IN, FILENAME_FORMAT, COLOR,
)
from ..layout		import Region, Text, Image, Box, Coordinate, layout_pdf
from .ethereum		import tokeninfo, tokenprices, tokenknown

"""
Invoice artifacts:

    Invoice/Quote	-- States deliverables, prices/taxes, and totals in various currencies

    Receipt		-- The Invoice, marked "PAID"

"""
log				= logging.getLogger( "artifact" )

# Custom tabulate format that provides "====" SEPARATING_LINE between line-items and totals
tabulate._table_formats["totalize"] = tabulate.TableFormat(
    lineabove		= tabulate.Line("", "-", "  ", ""),
    linebelowheader	= tabulate.Line("", "-", "  ", ""),
    linebetweenrows	= tabulate.Line("", "=", "  ", ""),
    linebelow		= tabulate.Line("", "-", "  ", ""),
    headerrow		= tabulate.DataRow("", "  ", ""),
    datarow		= tabulate.DataRow("", "  ", ""),
    padding		= 0,
    with_header_hide	= ["lineabove", "linebelow", "linebetweenrows"],
)
tabulate.multiline_formats["totalize"]	= "totalize"


def tabulate_nopad( *args, **kwds ):
    """Removes the default 2-character tabulate.MIN_PADDING from around column headers."""
    try:
        min_padding	= tabulate.MIN_PADDING
        tabulate.MIN_PADDING = 0
        return tabulate.tabulate( *args, **kwds )
    finally:
        tabulate.MIN_PADDING = min_padding


# An invoice line-Item contains the details of a component of a transaction.  It is priced in a
# currency, with a number of 'units', and a 'price' per unit.
#
# The 'tax' is a proportion of the computed total amount allocated to taxation; if <1 (eg. 0.05 or
# 5%), then it is added to the total amount; if > 1 (eg. 1.05 or 5%), then the prices is assumed to
# be tax-inclusive and it is subtracted from the computed amount.
#
# Each line-Item may be expressed in a different currency.
Item				= namedtuple( 'Item', [
    'description',			# "Widgets for The Thing"
    'units',				# 198
    'price',				# 1.98 eg. $USD, Fraction( 10000, 12 ) * 10**9, eg. ETH Gwei per dozen, in Wei
    'tax',				# 0.05, Fraction( 5, 100 ) eg. GST, added to amount, 1.05 GST, included in amount
    'decimals',				# Number of decimals to display computed amounts, eg. 2.  Default: currency's decimals//3
    'currency'				# USD
] )


@dataclass
class Item:
    description: str				# "Widgets for The Thing"
    price: Union[int,float,Fraction]		# 1.98 eg. $USDC, Fraction( 10000, 12 ) * 10**9, eg. ETH Gwei per dozen, in Wei
    units: Union[int,float]	= 1		# 198, 1.5
    tax: Optional[float,Fraction] = None        # 0.05, Fraction( 5, 100 ) eg. GST, added to amount, 1.05 GST, included in amount
    decimals: Optional[int]	= None		# Number of decimals to display computed amounts, eg. 2.; default is 1/3 of token decimals
    currency: Optional[str]	= None		# Default: $USD (value proxy is USDC)


class LineItem( Item ):
    def net( self ):
        """Computes the LineItem total 'amount', the 'taxes', and info on tax charged.  Use the
        full-precision percentage, to support fractional tax percentages.

        Uses the default Invoice Currency if none specified.

        """
        amount			= self.units * self.price
        taxes			= 0
        taxinf			= 'no tax'
        if self.tax:
            if self.tax < 1:
                taxinf		= f"{float( self.tax * 100 ):g}% add"
                taxes		= amount * self.tax
                amount	       += taxes
            elif self.tax > 1:
                taxinf		= f"{float(( self.tax - 1 ) * 100 ):g}% inc"
                taxes		= amount - amount / self.tax
        return amount, taxes, taxinf  # denominated in self.currencies


def conversions_table( conversions, symbols=None, greater=None, tablefmt=None, precision=None ):
    """Output a tabulation of conversion ratios, optionally >= a value (or of a pair of ratios), to
    the specified precision (default, 8 decimal points).

    Sort by the column with the highest density of conversion ratios.

    """
    if symbols is None:
        symbols			= sorted( set( sum( conversions.keys(), () )))
    if greater is None:
        greater			= .001
    if precision is None:
        precision		= 8
    symbols			= list( symbols )

    # The columns are typed according to the least generic type that *all* rows are convertible
    # into.  So, if any row is a string, it'll cause the entire column to be formatted as strings.
    def fmt( v, d ):
        if v:
            return round( v, d )
        return v

    headers_raw			= [ 'Coin' ] + [ f"in {s}" for s in symbols ]
    convers_raw			= [
        [ r ] + list(
            (
                '' if r == c or (r,c) not in conversions
                else '' if ( greater and conversions.get( (r,c) ) and conversions.get( (c,r) )
                             and ( conversions[r,c] < ( greater if isinstance( greater, (float,int) ) else conversions[c,r] )))
                else fmt( conversions[r,c], 8 )   # ie. to the Sat (1/10^8 Bitcoin)
            )
            for c in symbols
        )
        for r in symbols
    ]
    # Now, remove any column that are completely blank ('', not None).  This will take place for
    # worthless (or very low valued) currencies, esp. w/ a small 'greater'.  Transpose the
    # conversions (so each column is a row), and elide any w/ empty conversions rows.  Finally,
    # re-transpose back to columns.
    convers_txp			= list( zip( *convers_raw ))
    headers_use,convers_use_txp	= zip( *sorted(
        [
            (hdr,col)
            for hdr,col in zip( headers_raw, convers_txp )
            if any( c != '' for c in col )
        ],
        key	= lambda hdr_col: sum( map( bool, hdr_col[1] )),  	# number of occupied entries in column
        reverse	= True,
    ))
    # Finally sort by the 2nd (first non-label) column, which has the most occupied values; consider None/'' entries as zero
    convers_use			= sorted(
        zip( *convers_use_txp ),
        key	= lambda row: row[1] or 0,				# by 1st numeric column's conversion ratio
        reverse	= True,
    )

    return tabulate_nopad(
        convers_use,
        headers		= headers_use,
        floatfmt	= f',.{precision}g',
        intfmt		= ',',
        missingval	= '?',
        tablefmt	= tablefmt or INVOICE_FORMAT,
    )


class ConversionError( Exception ):
    pass


def conversions_remaining( conversions, verify=None ):
    """Complete the graph of conversion ratios, if we have a path from one pair to another.  Returns
    falsey if no additional conversion ratios were computable; truthy if some *might* be possible.
    Each run looks for single-hop conversions available.

    Put any desired conversions (eg. DOGE/USD) into conversions w/ None as value.

    For example, if ETH/USD: 1234.56 and BTC/USD: 23456.78, then we can deduce BTC/ETH:
    19.0001134007 and ETH/BTC: 0.05263126482.  Then, on the next call, we could compute DOGE/USD ==
    0.090308 if we provide BTC/DOGE: 3.85e-6

    Updates the supplied { ('a','b'): <ratio>, ...} dict, in-place.

    If NO None values remain after all computable ratios are deduced, returns False, meaning "no
    remaining unresolved conversions".

    Otherwise, return None iff we couldn't deduce any more conversion ratios, but there remain some
    unresolved { (a,b): None, ...} in conversions.

    """
    updated			= False
    # First, take care of any directly available one-hop conversions.
    for (a,b),r in list( conversions.items() ):
        if r and conversions.get( (b,a) ) is None:
            conversions[b,a]	= 1/r
            log.info( f"Deduced  {b:>6}/{a:<6} = {float( 1/r ):13.6f} from {a:>6}/{b:<6} == {float( r ):13.6f} == {r}" )
            updated		= True
        if r is None:
            continue
        for (a2,b2),r2 in list( conversions.items() ):
            if r2 is None:
                continue
            if b == b2 and a != a2 and conversions.get( (a,a2) ) is None:
                if r == r2:
                    # A special case for zero-valued (or other identically valued) tokens: the ratio is 1
                    # Eg. ZEENUS/ETH=0/1 / WEENUS/ETH=0/1 --> ZEENUS/WEENUS=1/1
                    conversions[a,a2] = 1
                    log.info( f"Unity    {a:>6}/{a2:<6} = {float( conversions[a,a2] ):13.6f} from {a:>6}/{b:<6} == {float( r ):13.6f} / {a2:>6}/{b2:<6} == {float( r2 ):13.6f}" )
                    updated	= True
                elif r2:
                    # Eg. USD/BTC=25000/1 / DOGE/BTC=275000/1 --> USD/DOGE=1/4
                    conversions[a,a2] = r / r2
                    log.info( f"Divide   {a:>6}/{a2:<6} = {float( conversions[a,a2] ):13.6f} from {a:>6}/{b:<6} == {float( r ):13.6f} / {a2:>6}/{b2:<6} == {float( r2 ):13.6f}" )
                    updated	= True
            if b == a2 and a != b2 and conversions.get( (a,b2) ) is None:
                # Eg. USD/BTC=25000/1 * BTC/DOGE=1/275000 --> USD/DOGE=1/4
                conversions[a,b2] = r * r2
                log.info( f"Multiply {a:>6}/{b2:<6} = {float( conversions[a,b2] ):13.6f} from {a:>6}/{b:<6} == {float( r ):13.6f} x {a2:>6}/{b2:<6} == {float( r2 ):13.6f}" )
                updated		= True
    if updated:
        return True
    # OK, got all available a/b --> b/a and a/b * b/c --> a/c and a/b / c/b --> a/c.  See if we
    # can find any routes between the desired a/b pairs.
    for (a,b),r in conversions.items():
        if r is not None:
            continue
        for c2,r2 in conversions.items():
            if r2 and a in c2:
                for c3,r3 in conversions.items():
                    if c2 != c3 and r3 and b in c3 and ( x_s := set( c2 ).intersection( set( c3 ))):
                        # We want a/b and we have found a/x or x/a, and b/x or x/b.  Assume their
                        # inverse is in conversions, and is (also) non-zero.
                        log.warning( f"Found intersection between {c2!r} and {c3!r}: {x_s!r}" )
                        x,	= x_s
                        conversions[a,b] = conversions[a,x] * conversions[b,x]
                        log.info( f"Compute  {a:>6}/{b2:<6} = {float( conversions[a,b] )} from {a:>6}/{x:<6} == {float( conversions[a,x] ):13.6f} and {b:>6}/{x:<6} == {float( conversions[b,x] ):13.6f}" )
                        conversions[b,a] = 1 / conversions[a,b]
                        return True

    # No more currency pairs are deducible from our present data; If no more are desired (contain
    # None), then we can return falsey (we're done), otherwise, raise an Exception or return truthy:
    # a description of what pairs remain undefined.
    remains			= commas( sorted( f'{a}/{b}' for (a,b),r in conversions.items() if r is None ), final='and' )
    if not remains:
        return False
    resolved			= commas( sorted( f'{a}/{b}' for (a,b),r in conversions.items() if r and r > 1 ), final='and' )
    msg				= f"Failed to find ratio(s) for {remains} via {resolved}"
    if verify:
        raise ConversionError( msg )
    log.debug( msg )
    return msg


def cryptocurrency_symbol( name, chain=None, w3_url=None, use_provider=None ):
    try:
        return Account.supported( name )
    except ValueError as exc:
        log.info( f"Could not identify currency {name!r} as a supported Cryptocurrency: {exc}" )
    # Not a known core Cryptocurrency; a Token?
    try:
        return tokeninfo( name, w3_url=w3_url, use_provider=use_provider ).symbol
    except Exception as exc:
        log.warning( f"Failed to identify currency {name!r} as an ERC-20 Token: {exc}" )
        raise


def cryptocurrency_proxy( name, decimals=None, chain=None, w3_url=None, use_provider=None ):
    """Return the named ERC-20 Token (or a known "Proxy" token, eg. BTC -> WBTC).  If not a
    Token/Proxy, then return any known Cryptocurrency matching the name.  In this case, the
    caller must (typically) already be informed of the currency value of such a Cryptocurrency by
    other means.

    """
    try:
        return tokeninfo( name, w3_url=w3_url, use_provider=use_provider )
    except Exception as exc:
        log.info( f"Could not identify currency {name!r} as an ERC-20 Token: {exc}" )
    # Not a known Token; a core known Cryptocurrency?
    return tokenknown( name, decimals=decimals )


class Invoice:
    """The totals for some invoice line items, in terms of some currencies, payable into some
    Cryptocurrency accounts.

    - Each account's native cryptocurrency symbol (at least) is reflected in computed invoice totals.
      - Eg. if ETH, BTC, XRP and DOGE Accounts are provided, we'll totalize into all 4.
    - Each currency must correspond to one cryptocurrency account (eg. Ethereum ERC-20s --> ETH account)
      - Eg. "US Dollar" (USDC) or "Wrapper Bitcoin" (WBTC) --> ETH, "Bitcoin" or BTC --> BTC
    - Each currency can either be supplied a ratio in conversions, or have a "Proxy" Token for an Oracle
      - Eg. "BTC" --> WBTC Token, "XRP" --> conversions[('XRP','BTC'): 0.00001797

    Currencies may be supplied as a recognized slip39.Account.supported( <name> ); full lower-case
    names in Account.CRYPTO_NAMES.keys(), upper-case symbols in Account.CRYPTO_NAMES.values().  For
    all other than BTC and ETH, a conversion ratio to some common currency (eg. USD) must be
    supplied in conversions.

    Otherwise, a ERC-20 token contract address or a token's symbol or full name (for known
    cryptocurrencies eg. "BTC", "ETH", or some known ERC-20 tokens (~top 100).  Presently, we only
    support obtaining price ratios for invoice cryptocurrencies that have a highly-liquid Ethereum
    ERC-20 "proxy", eg. ETH(WETH), BTC(WBTC).

    Can emit the line items in groups with summary sub-totals and a final total.

    Reframes the price of each line in terms of the Invoice's currencies (default: USD), providing
    "Total <SYMBOL>" and "Taxes <SYMBOL>" for each Cryptocurrency symbols.  NOTE: These are a
    *running* Total; the LineItem Amount is each line's total, in terms of that line's currency.

    Now that we have Web3 details, we can query tokeninfos and hence format currency decimals
    correctly.

    """
    def __init__(
        self, lines,
        accounts: Sequence[Account],		 # [ <Account>, ... ]   .crypto is symbol, eg. BTC, ETH, XRP
        currencies: Optional[List]	= None,	 # "USD" | [ "USD", "BTC" ] (first-most currencies/accounts is conversion "reference" currency)
        conversions: Optional[Dict]	= None,  # { ("USD","ETH"): 1234.56, ("USD","BTC"): 23456.78, ...}
        w3_url			= None,
        use_provider		= None,
    ):
        self.lines		= list( lines )
        self.w3_url		= w3_url
        self.use_provider	= use_provider

        # Collect all desired Invoice currencies; named, and those associated with supplied
        # accounts.  These are symbols, names, or ERC-20 token addresses.  Some may translate
        # into things we can get prices for via an off-chain Oracle via the Ethereum blockchain,
        # but some may not -- these must be supplied w/ a ratio in conversions, to at least one
        # token we *do* have the ability to get the value of.  This requires the caller to have
        # some kind of price feed or oracle of their own; it is recommended to use the 1inch
        # OffchainOracle instead, by sticking to the main Cryptocurrencies for which we have
        # real-time price proxies, eg. USD(USDC), BTC(WBTC), ETH(WETH).  If None (supplied or in
        # sequence), replace with default INVOICE_CURRENCY.
        if isinstance( currencies, (str, type(None)) ):
            currencies		= [ currencies ]
        currencies		= set( currencies )
        if None in currencies:
            currencies.remove( None )
            currencies.add( INVOICE_CURRENCY )
        log.info( f"Found {len( currencies )} Invoice currencies: {commas( currencies, final='and')}" )

        # Convert all known Crypto-currencies or Tokens to symbols (all upper-case).  Any
        # unrecognized as known Cryptos or Tokens will raise an Exception.  This effectively de-dups
        # all accounts/currencies, but doesn't substitute know "proxy" tokens (eg. BTC <-> WBTC)
        currencies		= set(
            cryptocurrency_symbol( c, w3_url=w3_url, use_provider=use_provider )
            for c in currencies
        )
        log.info( f"Found {len( currencies )} Invoice currency symbols: {commas( currencies, final='and')}" )

        # Associate any Crypto accounts provided w/ their core cryptos (adding any new ones from
        # accounts, to self.currencies).  After this every Invoice currency symbol is in
        # currencies_account.keys(), and every one with a provided account address is in
        # currencies_account[symbol]
        currencies_account	= { c: None for c in currencies }
        if accounts:
            for a in accounts:
                assert currencies_account.get( a.symbol ) is None, \
                    f"Duplicate accounts given for {a.symbol}: {a.address} vs. {currencies_account[a.symbol]}"
                currencies_account[a.symbol] = a
        currencies		= set( currencies_account )
        log.info( f"Found {len( currencies )} Invoice currency symbols + accounts: {commas( currencies, final='and')}" )

        # Associate any Crypto/Tokens w/ available ERC-20 proxies (eg. BTC -> WBTC), and any ERC-20s
        # w/ any supplied ETH account (eg.USDC, WBTC --> <ETH account>) Thus, if BTC was selected
        # and an Ethereum account provided, the buyer can pay in BTC to the Bitcoin account, or in
        # ETH or WBTC to the Ethereum account.  This will add the symbols for all Crypto proxies to
        # currencies_account.
        currencies_alias	= {}  # eg. { "WBTC": TokenInfo( "BTC", ... ), ... }
        currencies_proxy	= {}  # eg. { "BTC": TokenInfo( "WBTC", ... ), ... }
        for c in currencies:
            try:
                currencies_proxy[c] = alias = tokeninfo( c, w3_url=w3_url, use_provider=use_provider )
            except Exception as exc:
                # May fail later, if no conversions provided for this currency
                log.info( f"Failed to find proxy for Invoice currency {c}: {exc}" )
            else:
                # Yup; a proxy for a Crypto eg. BTC -> WBTC was found, or a native ERC-20 eg. USDC
                # was found; associate it with any ETH account provided.
                if eth := currencies_account.get( 'ETH' ):
                    currencies_account[currencies_proxy[c].symbol] = eth
                try:
                    known	= tokenknown( c )
                except Exception:
                    pass
                else:
                    # Aliases; of course, the native symbol is included in known aliases
                    currencies_alias[alias.symbol] = known
                    currencies_alias[known.symbol] = known
                    log.info( f"Currency {c}'s Proxy Symbol {alias.symbol} is an alias for {known.symbol}" )
        log.info( f"Found {len( currencies_proxy )} Invoice currency proxies: {commas( ( f'{c}: {p.symbol}' for c,p in currencies_proxy.items() ), final='and')}" )
        log.info( f"Added {len( set( currencies_account ) - currencies )} proxies: {commas( set( currencies_account ) - currencies, final='and' )}" )
        log.info( f"Alias {len( currencies_alias )} symbols to their native Crytocurrencies: {commas( ( f'{a}: {c.symbol}' for a,c in currencies_alias.items() ), final='and')}" )
        currencies		= set( currencies_account )

        # Find all LineItem.currency -> Invoice.currencies conversions required.  This
        # establishes the baseline conversions ratio requirements to convert LineItems to each
        # Invoice currency.  No prices are yet found.  Also, ensure that there is a route to
        # deduce the price of every provided cryptocurrency conversion.  For example, if the
        # invoice payment currency is XRP, and we have been provided the {('XRP','BTC'):
        # <ratio>}, how do we convert from every other currency to XRP?  It must occur via BTC.
        # Therefore, we must require a conversion from each currency found to a core currency; in
        # our case, ETH (since it is the default conversion ratio currency for our Ethereum
        # Oracle).
        line_currencies		= set(
            line.currency or INVOICE_CURRENCY
            for line in self.lines
        )
        # Finally, add all remaining Invoice and LineItem cryptocurrencies to currencies_proxy, by
        # their original names, symbols and names.  Now, every Invoice and LineItem currency (by
        # name and alias) is represented in currencies_proxy.
        for lc in line_currencies | currencies:
            proxy		= cryptocurrency_proxy( lc, w3_url=w3_url, use_provider=use_provider )
            for alias in set( (lc, proxy.name, proxy.symbol) ):
                if alias in currencies_proxy:
                    assert currencies_proxy[alias] == proxy, \
                        f"Incompatible LineItem.currency {lc!r} w/ currency {alias!r}: \n{proxy} != {currencies_proxy[alias]}"
                currencies_proxy[alias] = proxy
        log.info( f"Found {len( currencies_proxy )} Invoice and LineItem currencies: {commas( ( f'{n}: {p.symbol}' for n,p in currencies_proxy.items() ), final='and')}" )

        line_symbols		= set(
            cryptocurrency_symbol( lc )
            for lc in line_currencies
        )
        log.info( f"Found {len( line_symbols )} LineItem currencies: {commas( line_symbols, final='and')}" )

        # Now that we have all line-item, invoice and account Cryptocurrency symbols, we can
        # identify all the conversion we require to satisfy the Invoice.  Every LineItem
        # Cryptocurrency symbol must be convertible to each Invoice Cryptocurrency symbol.
        if conversions is None:
            conversions		= {}
        for ls in line_symbols:
            for c in currencies:
                if ls != c:
                    conversions.setdefault( (ls,c), None )  # eg. ('USDC','BTC'): None

        self.accounts		= accounts
        self.currencies		= currencies		# { "USDC", "BTC", ... }
        self.currencies_account	= currencies_account    # { "USDC": "0xaBc...12D", "BTC": "bc1...", ... }
        self.currencies_proxy	= currencies_proxy      # { "BTC": TokenInfo( "WBTC", ... ), ... }
        self.currencies_alias	= currencies_alias      # { "WBTC": TokenInfo( "BTC", ... ), ... }
        self.conversions	= conversions		# { ("BTC","ETH"): 14.3914, ... }
        self.created		= datetime.utcnow().astimezone( timezone.utc )
        self.resolved		= self.created

    def unsatisfied( self ):
        return set(
            sum(
                (
                    pair
                    for pair,ratio in self.conversions.items()
                    if ratio is None
                ),
                ()
            )
        ) - { 'ETH' }

    def resolve( self ):
        unsatisfied		= self.unsatisfied()
        if not unsatisfied:
            return

        for c in unsatisfied:
            if (c,'ETH') not in self.conversions:
                log.info( f"Require {c:>6}/ETH conversion ratio" )
                self.conversions[c,'ETH'] = None

        # Resolve all resolvable conversions (ie. from any supplied), see what's left
        while ( remaining := conversions_remaining( self.conversions ) ) and not isinstance( remaining, str ):
            if log.isEnabledFor( logging.DEBUG ):
                log.debug( f"Working: \n{conversions_table( self.conversions, greater=False )}" )
        log.info( f"{'Remaining' if remaining else 'Resolved'}:\n{conversions_table( self.conversions, greater=False )}\n{f'==> {remaining}' if remaining else ''}" )

        while remaining:
            # There are unresolved LineItem -> Invoice currencies.  We need to get a price ratio between
            # the LineItem currency, and at least one of the main Invoice currencies.  Since we know we're dealing
            # in Ethereum ERC-20 proxies, we'll keep getting ratios between currencies and ETH (the
            # default from tokenprices).
            candidates		= self.unsatisfied()
            for c in candidates:
                try:
                    (one,two,ratio), = tokenprices( c, w3_url=self.w3_url, use_provider=self.use_provider )
                except Exception as exc:
                    log.info( f"Ignoring candidate {c} for price deduction: {exc}" )
                    continue
                if self.conversions.get( (c,two.symbol) ) is None or self.conversions.get( (one.symbol,two.symbol) ) is None:
                    if self.conversions.get( (c,two.symbol) ) is None:
                        log.info( f"Updated  {c:>6}/{two.symbol:<6} to: {ratio}" )
                        self.conversions[c,two.symbol] = ratio
                    if self.conversions.get( (one.symbol,two.symbol) ) is None:
                        log.info( f"Updated  {one.symbol:>6}/{two.symbol:<6} to: {ratio}" )
                        self.conversions[one.symbol,two.symbol] = ratio
                    break
            else:
                # We must reject any zero-valued Cryptocurrencies from target currencies!  If the
                # caller selects eg. WEENUS as a payment currency, and we allow it, it might result
                # in an "infinite" payment in a zero-valued currency as a payment option.  We can
                # detect this, right here, because we won't be able to compute any conversion
                # ratio between a valuable cryptocurrency, and a zero-valued cryptocurrency.
                # Instead failing, remove it as a payment option!
                raise ConversionError( f"Failed to resolve conversion ratios {remaining}, using {len( candidates )} remaining candidates {commas( candidates, final='and' )}" )
            while ( remaining := conversions_remaining( self.conversions ) ) and not isinstance( remaining, str ):
                if log.isEnabledFor( logging.DEBUG ):
                    log.debug( f"Working: \n{conversions_table( self.conversions, greater=False )}" )
            log.info( f"{'Remaining' if remaining else 'Resolved'}:\n{conversions_table( self.conversions, greater=False )}\n{f'==> {remaining}' if remaining else ''}" )
        self.resolved		= datetime.utcnow().astimezone( timezone.utc )

    def decimals( self, currency ):
        info			= self.currencies_proxy[currency]
        if info.symbol in self.currencies_alias:
            info		= self.currencies_alias[info.symbol]
        return info.decimals

    def headers( self ):
        """Output the headers to use in tabular formatting of iterator.  By default, we'd recommend hiding any starting
        with an _ prefix (this is the default).

        Beware of making column labels w/ spaces; the 2nd half of some column labels is assumed to
        be a Cryptocurrency symbol for formatting purposes.

        """
        return (
            '_#',
            'Description',
            'Qty',
            'Price',
            '_Tax',
            'Tax%',
            'Taxes',
            '_Net',
            'Amount',
            '_Currency',
            'Coin',
            '_Decimals',
            '_Token',
        ) + tuple(
            f"_Total {currency}"
            for currency in sorted( self.currencies )
        ) + tuple(
            f"_Taxes {currency}"
            for currency in sorted( self.currencies )
        )

    def __iter__( self ):
        """Iterate of lines, computing totals in each Invoice.currencies

        Note that numeric values may be int, float or Fraction, and these are retained through
        normal mathematical operations, at full available precision (even if this is beyond the
        "decimals" precision of the underlying cryptocurrency).

        The number of decimals desired for each line is provided in "_Decimals".  The native
        number of decimals must be respected in this line's calculations.  For example, if eg.  a
        0-decimals Token is used, the Net, Tax and Amount are rounded to 0 decimals.  All
        rounding/display decimals should be performed at the final display of the data.

        The sums in various currencies converted from each line are not rounded; the currency
        ratio and hence sum is full precision, and is only rounded at the final use.

        """

        self.resolve()  # Resolve any yet unresolved conversions.  May raise Exception, if unresolvable.

        tot			= {
            c: 0. for c in self.currencies
        }
        tax			= {
            c: 0. for c in self.currencies
        }

        for i,line in enumerate( self.lines ):
            line_amount,line_taxes,line_taxinf = line.net()
            line_net		= line_amount - line_taxes
            # The line's ERC-20 Token (or a known proxy for the named Cryptocurrency), or a
            # TokenInfo for the specified known Cryptocurrency.  Always displays the underlying
            # native Cryptocurrency symbol (if known), even if a "Proxy" is specified, or used
            # for pricing.
            line_currency	= line.currency or INVOICE_CURRENCY  # Eg. "Bitcoin", "WBTC", "BTC"
            line_curr		= self.currencies_proxy[line_currency]
            log.info( f"Line currency {line_currency} has proxy {line_curr.symbol:6}: decimals: {line_curr.decimals}" )
            if line_curr.symbol in self.currencies_alias:
                line_curr	= self.currencies_alias[line_curr.symbol]
                log.info( f"Line currency {line_currency} alias for {line_curr.symbol:6}: decimals: {line_curr.decimals}" )
            line_symbol		= line_curr.symbol
            line_decimals	= line.decimals
            if line_decimals is None:
                line_decimals	= line_curr.decimals // 3

            for c in self.currencies:
                if line_curr.symbol == c:
                    tot[c]     += line_amount
                    tax[c]     += line_taxes
                else:
                    tot[c]     += line_amount * self.conversions[line_curr.symbol,c]
                    tax[c]     += line_taxes  * self.conversions[line_curr.symbol,c]
            yield (
                i,			# The LineItem #, and...
                line.description,
                line.units,
                line.price,
                line.tax,
                line_taxinf,
                line_taxes,
                line_net,
                line_amount,
                line_currency,		# the line's currency
                line_symbol,		# and Token symbol
                line_decimals,		# the desired number of decimals
                line_curr,		# and the associated tokeninfo
            ) + tuple(
                tot[c] for c in sorted( self.currencies )
            ) + tuple(
                tax[c] for c in sorted( self.currencies )
            )

    def pages(
        self,
        page		= None,
        rows		= None,
    ):
        """Yields a sequence of lists containing rows, paginated into 'rows' chunks.

        If page (zero-basis) specified, only yields matching pages.

        By default, we'll assume we need about 1/2 the rows for non-line-item invoice details
        (metadata, totals, conversions, ...)

        """
        if rows is None:
            rows		= INVOICE_ROWS // 2

        def page_match( i ):
            if page is None:
                return True
            if hasattr( page, '__contains__' ):
                return i in page
            return i == page

        page_i,page_l		= 0,[]
        for row in iter( self ):
            page_l.append( row )
            if len( page_l ) >= rows:
                if page_match( page_i ):
                    yield page_l
                page_i,page_l	= page_i+1,[]
        if page_l and page_match( page_i ):
            yield page_l

    def tables(
        self, *args,
        tablefmt	= None,
        columns		= None,		# list (ordered) set/predicate (unordered)
        totalfmt	= None,
        totalize	= None,		# Final page includes totalization
        description_max	= None,		# Default to some reasonable max
        **kwds				# page, rows, ...
    ):
        """Tabulate columns into textual tables.  Each page yields 3 items:

            <page>,<line-items>,<sub-totals>,<totals>

        The sub-totals are per-page; the totals are running totals of all pages produced thus far;
        hence the last page's totals are the full invoice totals.

        The 'columns' is a filter (a set/list/tuple or a predicate) that selects the columns
        desired; by default, elides any columns having names starting with '_'.  We will accept
        matches case-insensively, and ignoring any leading/trailing '_'.

        If any columns w/ leading '_' *are* selection, we trim the '_' for display.

        """
        if totalize is None:
            totalize		= True

        headers			= self.headers()

        def can( c ):
            """Canonicalize a header/column name for comparison"""
            return c.strip( '_' ).lower()

        headers_can		= [ can( h ) for h in headers ]
        if columns:
            if is_listlike( columns ):
                try:
                    selected	= tuple( headers_can.index( can( c )) for c in columns )
                except ValueError as exc:
                    raise ValueError( f"Columns not found: {commas( c for c in columns if can( c ) not in headers_can )}" ) from exc
            elif hasattr( columns, '__contains__' ):
                selected	= tuple( i for i,h in enumerate( headers ) if can( h ) in [ can( c ) for c in columns ] )
                assert selected, \
                    "No columns matched"
            else:
                # User-provided column predicate; do not canonicalize
                selected	= tuple( i for i,h in enumerate( headers ) if columns( h ) )
        else:
            # Default; just ignore _... columns
            selected		= tuple( i for i,h in enumerate( headers ) if not h.startswith( '_' ) )
        headers_selected	= tuple( headers[i] for i in selected )

        log.info( f"Tabulating indices {commas(selected, final='and')}: {commas( headers_selected, final='and' )}" )
        pages			= list( self.pages( *args, **kwds ))

        # Overall formatted decimals for the entire invoice; use the greatest desired decimals of
        # any token/line, based on Cryptocurrency proxies.  Individual line items may have been
        # rounded to lower decimals.
        #decimals_i		= headers_can.index( can( '_Decimals' ))
        #decimals		= max( line[decimals_i] for page in pages for line in page )
        #floatfmt		= ',f'  # f',.{decimals}f'
        #intfmt			= ','
        page_prev		= None
        for p,page in enumerate( pages ):
            first		= page_prev is None
            final		= p + 1 == len( pages )

            # {Sub}total for payment cryptocurrrencies are rounded to the individual
            # Cryptocurrency's designated decimals // 3.  For typical ERC-20 tokens, this is
            # eg. USDC: 6 // 3 == 2, WBTC: 18 // 3 == 6.  For known Cryptocurrencies, eg. BTC: 24 //
            # 3 == 8.
            #
            # If a symbol is a "proxy" token for some upstream known cryptocurrency, then the
            # upstream native cryptocurrency's decimals should be used, so that all lines match.  We
            # keep track of each cryptocurrency_alias[<proxy-symbol>] -> <original-symbol>
            #
            # TODO: There should be a more sensible / less brittle way to do this.
            def deci( c ):
                return self.decimals( c ) // 3

            def toti( c ):
                return headers_can.index( can( f'_Total {c}' ))

            def taxi( c ):
                return headers_can.index( can( f'_Taxes {c}' ))

            # The per-currency Sub-totals (for each page), sorted in ascending order.  Stabilizes
            # the sort by also sorting the currencies.
            subtotal_rows	= sorted(
                [
                    [
                        str( self.currencies_account[c] ),
                        round( page[-1][taxi( c )], deci( c )) if first else round( page[-1][taxi( c )] - page_prev[-1][taxi( c )], deci( c )),
                        round( page[-1][toti( c )], deci( c )) if first else round( page[-1][toti( c )] - page_prev[-1][toti( c )], deci( c )),
                        c,
                        self.currencies_account[c].name if c == self.currencies_account[c].symbol else self.currencies_proxy[c].name,
                    ]
                    for c in sorted( self.currencies )
                ],
                key	= lambda r: r[2]
            )
            subtotal_headers	= (
                'Account',
                'Taxes' if final else f'Taxes {p+1}/{len( pages )}',
                'Subtotal' if final else f'Subtotal {p+1}/{len( pages )}',
                'Coin',
                'Currency',
            )
            subtotal		= tabulate_nopad(
                subtotal_rows,
                headers		= subtotal_headers,
                intfmt		= ',',
                floatfmt	= ',.15g',
                tablefmt	= tablefmt or INVOICE_FORMAT,
            )

            # And the per-currency Totals (up to current page)
            total_rows		= sorted(
                [
                    [
                        str( self.currencies_account[c] ),
                        round( page[-1][taxi( c )], deci( c )),
                        round( page[-1][toti( c )], deci( c )),
                        c,
                        self.currencies_account[c].name if c == self.currencies_account[c].symbol else self.currencies_proxy[c].name,
                    ]
                    for c in sorted( self.currencies )
                ],
                key	= lambda r: r[2]
            )
            total_headers	= (
                'Account',
                'Taxes' if final else f'Taxes {p+1}/{len( pages )}',
                'Total' if final else f'Total {p+1}/{len( pages )}',
                'Coin',
                'Currency',
            )
            total		= tabulate_nopad(
                total_rows,
                headers		= total_headers,
                intfmt		= ',',
                floatfmt	= ',.15g',
                tablefmt	= tablefmt or INVOICE_FORMAT,
            )

            def fmt( val, hdr, coin  ):
                hdr_split	= hdr.split()  # Eg. "Total USDC" --> ["Total","USDC"]
                if can( hdr_split[0] ) in ( 'total', 'taxes', 'amount', 'net' ):
                    # It's a computed Total/Taxes/Amount/Net; use the line's/column's
                    # cryptocurrency to round.  Do not round Price, because multiples of a
                    # fractional price are significant, even if the precision is greater than the
                    # number of decimals supported by the LineItem's Cryptocurrency.
                    decimals	= deci( hdr_split[1] if len( hdr_split ) > 1 else coin )
                    val		= round( val, decimals )
                return val

            # Produce the page line-items.  We must round each line-item's numeric price values
            # according to the target coin.
            table_rows		= [
                [
                    fmt( line[i], h, line[headers_can.index( can( 'Coin' ))] )
                    for i,h in zip( selected, headers_selected )
                ]
                for line in page
            ]
            table_headers	= [ h.strip('_') for h in headers_selected ]

            if final and totalize:
                # Include a separator, followed by each cryptocurrency's totalization.  Map
                # total_rows columns:
                #
                #   Account  -> Description
                #   Total    -> Amount
                #   Coin     -> Coin
                #   Currency -> Currency
                table_rows.append( tabulate.SEPARATING_LINE )
                for acc,tax,tot,coin,curr in total_rows:
                    row		= []
                    for h in headers_selected:
                        if can( h ) == can( 'Description' ):
                            row.append( acc )
                        elif can( h ) == can( 'Amount' ):
                            row.append( tot )
                        elif can( h ) == can( 'Taxes' ):
                            row.append( tax )
                        elif can( h ) == can( 'Coin' ):
                            row.append( coin )
                        elif can( h ) == can( 'Currency' ):
                            row.append( curr )
                        else:
                            row.append( None )
                    table_rows.append( row )

            # Presently, the Description column is the only one likely to have a width issue...
            maxcolwidths		= None
            if description_max is None:
                description_max		= INVOICE_DESCRIPTION_MAX
            if description_max:
                desc_i			= headers_can.index( can( "Description" ))
                maxcolwidths	= [
                    description_max if i == desc_i else None
                    for i in selected
                ]
                log.info( f"Maximum col widths {commas( ( f'{h}: {w}' for h,w in zip( table_headers, maxcolwidths ) ), final='and' )}" )

            table		= tabulate_nopad(
                table_rows,
                headers		= table_headers,
                intfmt		= ',',
                floatfmt	= ',.15g',
                tablefmt	= tablefmt or INVOICE_FORMAT,
                maxcolwidths	= maxcolwidths,
            )

            yield page, table, subtotal, total


def layout_invoice(
    inv_dim: Coordinate,			# Printable invoice dimensions, in inches (net page margins).
    inv_margin: int,				# Additional margin around invoice
    rows: int,					# Compute rows, from the number of columns
):
    """Layout an Invoice, in portrait format.  Assumes that we are laying out an invoice in the
    printable area of a page (ie. net of non-printable page margins, eg. on a 8" x 10.5"/13.5"
    viewport onto a 8.5" x 11"/14" Letter/Legal page.

    Rotates the watermark, etc. so its angle is from the lower-left to the upper-right.

                 b
          +--------------+        +--------------+
          |             .         |I.. (img)  Na.|
          |           D.          |--------------|
          |          .            |# D.. u.u.    |
          |        I.             |...           |
          |       .               |         ---- |
          |     A. c              |    BTC  #.## |
          |    .                  |    ETH  #.## |
        a |  P.                   | (terms)      |
          |β.                     | (tax #, ..)  |
          + 90-β                  +--------------+

       tan(β) = b / a, so
       β = arctan(b / a)

    Sets priority:
      -3     : Hindmost backgrounds
      -2     : Things atop background, but beneath contrast-enhancement
      -1     : Contrast-enhancing partially transparent images
       0     : Text, etc. (the default)

    All sizes in mm.

    """
    prio_backing		= -3
    prio_normal			= -2
    prio_contrast		= -1

    if inv_margin is None:
        inv_margin		= 0		# Default; no additional margin (besides page margins)

    inv				= Box( 'invoice', 0, 0, inv_dim.x, inv_dim.y )
    inv_int			= inv.add_region_relative(
        Region( 'inv-interior', x1=+inv_margin, y1=+inv_margin, x2=-inv_margin, y2=-inv_margin )
    ).add_region_proportional(
        # inv-image	-- A full background image, out to the margins; bottom-most in stack
        Image(
            'inv-image',
            priority	= prio_backing,
        )
    )

    a				= inv_int.h
    b				= inv_int.w
    c				= math.sqrt( a * a + b * b )
    β				= math.atan( b / a )
    rotate			= 90 - math.degrees( β )

    # Header: Vendor name & contact on left, Invoice/Quote/Receipt and logo on right
    #           8"
    #     +-------------------------------------------------------------------------------+
    #     | Dominion R&D Corp                                                      INVOICE|
    #     | (vendor info)                                               //////////////////|
    # 2"  | Bill To:                  Invoice No: CLI-20231201-0001     ///16x9 LOGO//////|
    #     | Client Name             Invoice Date: ...                   //////////////////|
    #     | (client info)                    Due: ...                   //////////////////|
    #     + ...                                                         //////////////////+

    # inv-head: Header Image 2"x8" header of Invoice (if no full-page inv-image).  Since our length
    # may be variable for Letter/Legal, compute the length of the header as a fraction of its width.

    head			= inv_int.w * 1/4  / inv_int.h  # eg.   2" x 8"
    foot		    = 1 - inv_int.w * 1/16 / inv_int.h  # eg. 0.5" x 8"

    inv_head			= inv_int.add_region_proportional(
        Image(
            'inv-head',
            y2		= head,
            priority	= prio_normal,
        ),
    )

    # inv-vendor: Vendor name and Invoice label, top of invoice in large font.  The remaining
    # vertical portion of the header contains the 16/9 format Logo, positioned in the LR corner.
    logo			= 1/6				# 1/3" of 2" header for vendor/label
    logo_v_frac			= 1 - logo
    logo_h_frac			= 1 - inv_head.h * logo_v_frac * 16/9 / inv_head.w
    inv_head.add_region_proportional(
        Image(
            'inv-vendor-bg',
            x2		= logo_h_frac,
            y2		= logo,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'inv-vendor',
            align	= 'L',
        )
    )
    inv_head.add_region_proportional(
        Image(
            'inv-vendor-info-bg',
            y1		= logo,
            x2		= logo_h_frac,
            y2		= 1/2,
            priority	= prio_contrast,
        )
    ).add_region_proportional(
        Text(
            'inv-vendor-info',
            y2		= 1/4,   # 3 lines in remaining upper half (2/6) of header
            multiline	= True,
        )
    )

    # inv-client[-info] In lower 1/2 of LHS of header
    inv_head.add_region_proportional(
        Image(
            'inv-client-bg',
            y1		= 1/2,
            y2		= 1/2 + logo,
            x2		= logo_h_frac,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'inv-client',
            align	= 'L',
            italic	= True,
        )
    )
    inv_head.add_region_proportional(
        Image(
            'inv-client-info-bg',
            y1		= 1/2 + logo,
            x2		= logo_h_frac,
            priority	= prio_contrast,
        )
    ).add_region_proportional(
        Text(
            'inv-client-info',
            y2		= 1/4,   # 4 lines in remaining lower half (2/6) of header
            multiline	= True,
            italic	= True,
        )
    )

    # inv-label, inv-logo: Label "Invoice" and 16/9 logo in LR corner
    inv_head.add_region_proportional(
        Image(
            'inv-label-bg',
            x1		= logo_h_frac,
            y2		= logo,
            priority	= prio_contrast,
        )
    ).add_region(
        Text(
            'inv-label',
            align	= 'R'
        )
    )
    inv_head.add_region_proportional(
        Image(
            'inv-logo',
            x1		= logo_h_frac,
            y1		= logo,
            priority	= prio_normal,
        ),
    )

    # inv-body, ...: Image for Body of Invoice;
    # inv-table: Main Invoice text area
    inv_body			= inv_int.add_region_proportional(
        Image(
            'inv-body',
            y1		= head,
            y2		= foot,
            priority	= prio_normal,
        )
    )

    inv_body.add_region_proportional(
        Image(
            'inv-table-bg',
            priority	= prio_contrast,
        ),
    ).add_region_proportional(
        Text(
            'inv-table',
            y2		= 1/rows,
            font	= 'mono',
            #size_ratio	= 9/16,
            multiline	= True,
            bold	= True,
        )
    )

    # inv-foot: Image for bottom of Invoice (if no inv-image)
    inv_int.add_region_proportional(
        Image(
            'inv-foot',
            y1		= foot,
            priority	= prio_normal,
        ),
    )

    # Finally, in front of all other "contrast" images, put the watermark
    inv_int.add_region_proportional(
        Text(
            'inv-watermark',
            x1		= c/b * 10/100,
            y1		= +0/32,
            x2		= c/b * 90/100,
            y2		= +8/32,
            foreground	= int( COLOR[-2], 16 ),  # med grey
            rotate	= -rotate,
            bold	= True,
            italic	= True,
            align	= 'C',
            priority	= prio_contrast,
            text	= "- Sample -",
        )
    )

    return inv


def datetime_advance( dt, years=None, months=None, days=None, hours=None, minutes=None, seconds=None ):
    """Advance a datetime the specified amount.  Differs from timedelta, in that A) months are supported
    (so we must compute the target month, and clamp the day number appropriately).  Otherwise, uses
    timedelta.

    """
    if years or months:
        yr,mo,dy		= dt.year,dt.month,dt.day
        yr		       += years or 0
        mo		       += months or 0
        if mo > 12:
            yr		       += ( mo - 1 ) // 12
            mo			= ( mo - 1 ) % 12 + 1
            assert 1 <= mo <= 12
        _,dy_max		= monthrange( yr, mo )
        if dy > dy_max:
            dy			= dy_max
        dt			= dt.replace( day=1 ).replace( year=yr ).replace( month=mo ).replace( day=dy )
    if days or hours or minutes or seconds:
        dt		       += timedelta( days=days or 0, hours=hours or 0, minutes=minutes or 0, seconds=seconds or 0 )
    return dt


@dataclass
class Contact:
    name: str					# Company or Individual eg. "Dominion Research and Development Corp."
    contact: str				# client/vendor authority eg. "Perry Kundert <perry@dominionrnd.com>"
    phone: Optional[str]	= None
    address: Optional[str]	= None		# multi-line mailing/delivery address
    billing: Optional[str]	= None		# and billing address, if different

    @property
    def info( self ):
        if self.contact:
            yield self.contact
        if self.phone:
            yield self.phone
        if self.address:
            yield ', '.join( filter( None, self.address.split( '\n' )))
        if self.billing:
            yield "Billing: " + ', '.join( filter( None, self.billing.split( '\n' )))


@dataclass
class InvoiceMetadata:
    """Collected Invoice metadata, required to generate an Invoice.  An identical Invoice can be
    issued with multiple different InvoiceMetadata, eg. to different clients (optionally may be
    absent, indicating a generic invoice to an anonymous client).

    """
    vendor: Contact				# Vendor's identifying info
    client: Optional[Contact] = None  		# Client's identifying info (eg. Name, Attn, Address)
    number: Optional[Union[str,int]] = None     # eg. "CLI-20230930-0001" (default: <client>-<date>-<num>)
    date: Optional[datetime] = None		# default: now
    due: Optional[datetime] = None		# default: 1 month
    terms: Optional[str] = None			# default: Net Duration(due - date)
    directory: Optional[Union[str,Path]] = None  # Location of assets (and invoices) (./)
    label: Optional[str] = None			# eg. Invoice


@dataclass
class InvoiceOutput:
    invoice: Invoice
    metadata: InvoiceMetadata
    pdf: fpdf.FPDF
    path: Optional[Path] = None			# optionally written to path


def produce_invoice(
    invoice: Invoice,				# Invoice LineItems, Cryptocurrency Account data
    metadata: InvoiceMetadata,			# vendor, client, number, etc.
    rows: Optional[int]		= None,
    paper_format: Any		= None, 	# 'Letter', 'Legal', 'A4', (x,y) dimensions in mm.
    orientation: Optional[str]	= None,		# available orientations; default portrait, landscape
    image: Optional[Union[Path,str]] = None,  # A custom 8"/10.5" full background image (Path/name relative to directory/'.'
    logo: Optional[Union[Path,str]] = None,  # A custom 16/9 logo image (Path or name relative to directory/'.'
):
    """Produces a PDF containing the supplied Invoice details, optionally with a PAID watermark.

    """
    if ( label := metadata.label ) is None:
        label			= 'Invoice'
    if isinstance( directory := metadata.directory, (str,type(None))):
        directory		= Path( directory or '.' )
    assert isinstance( directory, (Path,type(None)))
    log.info( f"Dir.:  {directory}" )

    if isinstance( image, (str,type(None))):
        image			= directory / ( image or 'inv-image.png' )
        if not image.exists():
            image		= None
    assert isinstance( image, (Path,type(None)))
    log.info( f"Image: {image}" )

    if isinstance( logo, (str,type(None))):
        logo			= directory / ( logo or 'inv-logo.png' )
        if not logo.exists():
            logo		= None
    assert isinstance( logo, (Path,type(None)))
    log.info( f"Logo:  {logo}" )

    # Any datetime WITHOUT a timezone designation is re-interpreted as the local timezone of the
    # invoice issuer, or UTC.
    if ( date := metadata.date ) is None:
        date			= invoice.created
    if date.tzname() is None:
        try:
            zone		= get_localzone()
        except Exception:
            zone		= timezone.utc
        date			= date.astimezone( zone )

    log.info( f"Date:  {date.strftime( INVOICE_STRFTIME )}" )
    if not isinstance( due := metadata.due, datetime ):
        if not due:
            due			= INVOICE_DUE
        log.info( f"Due w/ {due!r}" )
        if is_mapping( due ):
            due			= datetime_advance( date, **dict( due ))
        elif is_listlike( due ):
            due			= datetime_advance( date, *tuple( due ))
        elif isinstance( due, timedelta ):
            due			= date + due
        else:
            raise ValueError( f"Unsupported Invoice due date: {due!r}" )
    log.info( f"Due:   {due.strftime( INVOICE_STRFTIME )}" )
    if ( terms := metadata.terms ) is None:
        terms			= f"Net {Duration( due - date )!r}"
    log.info( f"Terms: {terms}" )

    if ( number := metadata.number ) is None:
        cli			= metadata.client.name.split()[0].upper()[:3] if metadata.client else 'INV'
        ymd			= date.strftime( '%Y%m%d' )
        key			= f"{cli}-{ymd}"
        produce_invoice.inv_count[key] += 1
        num			= produce_invoice.inv_count[key]
        number			= f"{cli}-{ymd}-{num:04d}"
    log.info( f"Num.:  {number}" )

    # We can now capture the actual metadata used for this specific invoice
    metadata			= InvoiceMetadata(
        vendor		= metadata.vendor,
        client		= metadata.client,
        number		= number,
        date		= date,
        due		= due,
        terms		= terms,
        directory	= directory,
        label		= label,
    )

    # Default to full page, given the desired paper and orientation.  All PDF dimensions are mm.
    invs_pp,orientation,page_xy,pdf,comp_dim = layout_pdf(
        paper_format	= paper_format,
        orientation	= orientation,
    )
    log.info( f'Dim.: {comp_dim.x / MM_IN:6.3f}" x {comp_dim.y / MM_IN:6.3f}"' )

    # TODO: compute rows based on line lengths; the longer the line, the smaller the lines
    # required In the mean time, assume we can only fit about 1/2 the number of line-item rows,
    # as the invoice has total lines, due to the details, totals and currency conversion ratios.
    if rows is None:
        rows			= INVOICE_ROWS

    # Compute the Invoice layout on the page.  All page layouts are specified in inches.
    inv_dim			= Coordinate( comp_dim.x / MM_IN, comp_dim.y / MM_IN )

    inv				= layout_invoice( inv_dim=inv_dim, inv_margin=0, rows=rows )
    inv_elements		= list( inv.elements() )
    inv_tpl			= fpdf.FlexTemplate( pdf, inv_elements )

    p_cur			= None
    details			= list( invoice.tables() )
    for i,(page,tbl,sub,tot) in enumerate( details ):
        p,(offsetx,offsety)		= page_xy( i )
        if p != p_cur:
            pdf.add_page()

        here			= Path( __file__ ).resolve().parent
        layout			= here.parent / 'layout'
        #crypto			= here / 'Crypto'

        final			= i + 1 == len( details )
        inv_tpl['inv-image']	= image
        inv_tpl['inv-logo']	= logo
        inv_tpl['inv-label']	= f"{metadata.label} (page {p+1}/{len( details )})"
        inv_tpl['inv-label-bg']	= layout / '1x1-ffffffbf.png'

        inv_tpl['inv-vendor']	= metadata.vendor.name
        inv_tpl['inv-vendor-bg'] = layout / '1x1-ffffffbf.png'
        inv_tpl['inv-vendor-info'] = '\n'.join( metadata.vendor.info )
        inv_tpl['inv-vendor-info-bg'] = layout / '1x1-ffffffbf.png'

        if metadata.client:
            inv_tpl['inv-client'] = "Bill To: " + metadata.client.name
            inv_tpl['inv-client-bg'] = layout / '1x1-ffffffbf.png'
            inv_tpl['inv-client-info'] = '\n'.join( metadata.client.info )
            inv_tpl['inv-client-info-bg'] = layout / '1x1-ffffffbf.png'

        dets			= tabulate_nopad(
            [
                [ f'{metadata.label} #:', metadata.number ],
                [ 'Date:', metadata.date.strftime( INVOICE_STRFTIME ) ],
                [ 'Due:', metadata.due.strftime( INVOICE_STRFTIME ) ],
                [ 'Terms:', metadata.terms ]
            ],
            colalign=( 'right', 'left' ), tablefmt='plain'
        )

        exch			= conversions_table( invoice.conversions )
        exch_date		= invoice.resolved.strftime( INVOICE_STRFTIME )
        inv_table		= (dets, tbl,)
        if final:
            inv_table	       += ( f"CONVERSION RATIOS (est. {exch_date}):\n{exch}", )
        inv_tpl['inv-table']	= '\n\n'.join( inv_table )
        inv_tpl['inv-table-bg']	= layout / '1x1-ffffffbf.png'

        inv_tpl.render( offsetx=offsetx, offsety=offsety )
    # Caller already has the Invoice; return the PDF and computed InvoiceMetadata
    return (paper_format,orientation),pdf,metadata
produce_invoice.inv_count	= defaultdict( int )  # noqa: E305


def write_invoices(
    invoices: Sequence[Tuple[Invoice,InvoiceMetadata]],  # sequence of [ (Invoice, InvoiceMetadata), ... ] or { "<name>": Invoice, ... }
    filename		= True,		# A file name/Path, if PDF output to file is desired; ''/True implies default., False no file
    **kwds
) -> Union[InvoiceOutput, Exception]:
    """Generate unique cryptocurrency account(s) for each client invoice, generate the invoice,
    and (optionally) write them to files.  Yields a sequence of the generated invoice PDF names
    and details.

    If an Exception is raised during Invoice generation, a (None, <Exception>) is generated,
    instead of a ("name", <InvoiceOutput).

    """
    if filename is None:
        filename		= True
    for invoice,metadata in invoices:  # Provides the supplied invoice,metadata...
        try:
            # and receives the transformed (specialized) metadata.
            _,pdf,metadata		= produce_invoice(
                invoice	= invoice,
                metadata = metadata,
                **kwds
            )

            accounts		= invoice.accounts
            assert accounts, \
                "At least one Cryptocurrency account must be specified"

            name		= (( '' if filename is True else filename ) or FILENAME_FORMAT ).format(
                name	= metadata.number,
                date	= datetime.strftime( metadata.date, '%Y-%m-%d' ),
                time	= datetime.strftime( metadata.date, '%H.%M.%S'),
                crypto	= accounts[0].crypto,
                address	= accounts[0].address,
            )
            if not name.lower().endswith( '.pdf' ):
                name		       += '.pdf'

            path			= None
            if filename is not False:
                path		= metadata.directory.resolve() / name
                log.warning( f"Writing Invoice {metadata.number!r} to: {path}" )
                pdf.output( path )

            output		= InvoiceOutput(
                invoice	= invoice,
                metadata = metadata,
                pdf	= pdf,
                path	= path,
            )
        except (BaseException, ConversionError) as exc:
            name, output	= None, exc

        yield name, output
